import pytest
from hypothesis import given
from hypothesis.strategies import integers

from stylo.time import Timeline


time = integers(min_value=1)


@pytest.mark.time
class TestTimeline:
    """Tests for the :code:`Timeline` class"""

    def test_init_defaults(self):
        """Ensure that sensible defaults are chosen if no arguments given."""

        timeline = Timeline()
        assert timeline.framecount == 0
        assert timeline.duration == 0
        assert timeline.fps == 25

    @given(seconds=time)
    def test_duration_is_framecount(self, seconds):
        """Ensure that when the frames per second is one, that the framecount matches
        the duration."""

        timeline = Timeline(seconds=seconds, fps=1)
        assert timeline.framecount == seconds
        assert timeline.duration == seconds

    @given(hours=time)
    def test_duration_calc_with_hours(self, hours):
        """Ensure that we can specify the duration in hours."""

        timeline = Timeline(hours=hours)
        assert timeline.duration == hours * 60 * 60

    @given(minutes=time)
    def test_duration_calc_with_minutes(self, minutes):
        """Ensure that we can specifiy the duration in minutes."""

        timeline = Timeline(minutes=minutes)
        assert timeline.duration == minutes * 60

    @given(seconds=time)
    def test_duration_calc_with_seconds(self, seconds):
        """Ensure that we can specify the duration in seconds."""

        timeline = Timeline(seconds=seconds)
        assert timeline.duration == seconds

    def test_build_filename(self):
        """Ensure that we can generate appropriate filnames to use to save image
        sequences."""

        timeline = Timeline(seconds=10)
        timeline._build_filename("my/animation/frame.png")

        assert timeline._filename == "my/animation/frame-{:03d}.png"

    def test_timeline_main(self):
        """Ensure that we can declare a main function and render it."""

        timeline = Timeline(seconds=10)

        class Counter:
            def __init__(self):
                self.count = 0

        counter = Counter()

        @timeline.main
        def animation(f, t):
            counter.count += 1

        timeline.render()
        assert counter.count == 250



def lerp(a, b):
    """Return a function that will linearly interpolate between :code:`a` and :code:`b`.

    The returned function will take the value :code:`a` when passed the value :code:`0`
    and will increase/decrease linearly to the value :code:`b`, taking that value at the
    point :code:`1`.

    :param float a: The value the function should take at :code:`0`
    :param float b: The value the function should take at :code:`1`
    """

    def s(t):
        return (1 - t) * a + t * b

    return s



import numpy.testing as npt
import pytest
from hypothesis import given
from math import pi
from unittest import TestCase

from stylo.domain import UnitSquare
from stylo.domain.transform.rotation import Rotation
from stylo.testing.domain import BaseRealDomainTest
from stylo.testing.strategies import dimension, angle


@pytest.mark.domain
class TestRotatation(TestCase, BaseRealDomainTest):
    """Tests for the translation class."""

    def setUp(self):

        self.domain = Rotation(UnitSquare(), pi / 2)

    @given(width=dimension, height=dimension, angle=angle)
    def test_angle_affects_t_values(self, width, height, angle):
        """Ensure that the :code:`angle` shifts each value by the negative of the value
        given by the user so that the rotation is intuitive. Also ensure that only the
        :code:`t` values are affected."""

        domain = UnitSquare()
        rs, ts = domain["rt"](width, height)

        rotated = Rotation(domain, angle)
        r_rs, r_ts = rotated["rt"](width, height)

        # The r values should be unaffected
        npt.assert_array_equal(rs, r_rs)

        difference = ts - r_ts
        npt.assert_almost_equal(difference, -angle)



import numpy as np

from stylo.color import RGB8
from stylo.domain import get_real_domain
from stylo.image import Image
from stylo.image.image import Drawable, render_drawable


class LayeredImage(Image):
    def __init__(self, background=None, scale=2, colorspace=None):
        self.scale = scale

        if background is None:
            background = "ffffff"

        if colorspace is None:
            colorspace = RGB8

        self.background = colorspace.parse(background)
        self.colorspace = colorspace
        self.layers = []

    def add_layer(self, shape, color, domain=None):

        # Make sure everyone uses the same colorspace.
        color.colorspace = self.colorspace

        self.layers.append(Drawable(domain, shape, color))

    def _render(self, width, height):

        domain = get_real_domain(width, height, self.scale)
        dimensions = (height, width, len(self.background))
        image_data = np.full(dimensions, self.background, dtype=np.uint8)

        for drawable in self.layers:

            if drawable.domain is None:
                drawable.domain = domain

            image_data = render_drawable(drawable, image_data)

        return image_data



from stylo.color.colormap import ColorMap


class FillColor(ColorMap):
    def __init__(self, color=None, colorspace=None):
        super().__init__(colorspace)

        if color is None:
            color = "000000"

        self.color = self._parse_color(color)

    def _paint(self, shape, image_data):

        image_data[shape] = self.color
        return image_data



from uuid import uuid4
from inspect import signature


def bounded_property(
    name,
    bounded_above=None,
    bounded_above_by=None,
    bounded_below=None,
    bounded_below_by=None,
):
    """Factory function to define a bounded property.

    This function writes a property definition for a numeric bounded property.
    It can be bounded above, below or both and the bound can either be a given
    number or another attribute/property on the same class.

    The bound can be specified by a constant numeric value e.g. :code:`1` or
    :code:`2.4` in which case you use the :code:`bounded_above` and
    :code:`bounded_below` arguments.

    Alternatively to specify the bound as the value of an attribute on the
    same class then use the :code:`bounded_above_by` and :code:`bounded_below_by`
    arugments and pass in the name of the attribute as a string.

    .. note::

       You cannot simultaneously use the :code:`bounded_xxx` and
       :code:`bounded_xxx_by` version of an argument.

    :param name: The name of the property
    :param bounded_above: The value to bound the property above by. This cannot
                          be used in conjunction with :code:`bounded_above_by`
    :param bounded_below: The value to bound the property below by. This cannot
                          be used in conjunction with :code:`bounded_below_by`
    :param bounded_above_by: The name of the attribute to bound the value by. This
                             cannot be used in conjunction with :code:`bounded_above`
    :param bounded_below_by: The name of the attribute to bound the value by. This
                             cannot be used in conjunction with :code:`bounded_below`.

    :type name: str
    :type bounded_above: float, int
    :type bounded_below: float, int
    :type bounded_below_by: str
    :type bounded_above_by: str

    :raises ValueError: If the given arguments are inconsistent in some way.
    :raises TypeError: If the given arguments do not have their expected types

    :returns: The constructed property instance
    """

    # First check that the given arguments make sense.
    if bounded_below is not None and bounded_below_by is not None:
        raise ValueError('You can only use "bounded_below" or "bounded_below_by"')

    if bounded_above is not None and bounded_above_by is not None:
        raise ValueError('You can only use "bounded_above" or "bounded_above_by"')

    if bounded_below is not None and not isinstance(bounded_below, (int, float)):
        raise TypeError("The value of bounded_below must be a number.")

    if bounded_above is not None and not isinstance(bounded_above, (int, float)):
        raise TypeError("The value of bounded_above must be a number.")

    if bounded_below_by is not None and not isinstance(bounded_below_by, (str,)):
        raise TypeError("The value of bounded_below_by must be a string")

    if bounded_above_by is not None and not isinstance(bounded_above_by, (str,)):
        raise TypeError("The value of bounded_above_by must be a string")

    hidden_name = "_" + name

    type_error_message = "Value of property {0} must be a number".format(name)
    bounded_below_message = "Value of property {0} must be strictly larger than {1}"
    bounded_below_by_message = (
        "Value of property {0} must be strictly larger than the value of property {1}"
    )
    bounded_above_message = "Value of property {0} must be strictly less than {1}"
    bounded_above_by_message = (
        "Value of property {0} must be strictly less than the value of property {1}"
    )

    checks = [(lambda s, v: isinstance(v, (int, float)), TypeError(type_error_message))]

    if bounded_above is not None:
        exception = ValueError(bounded_above_message.format(name, bounded_above))
        checks.append((lambda s, v: v < bounded_above, exception))

    if bounded_above_by is not None:
        exception = ValueError(bounded_above_by_message.format(name, bounded_above_by))
        checks.append(
            (lambda s, v: v < s.__getattribute__(bounded_above_by), exception)
        )

    if bounded_below is not None:
        exception = ValueError(bounded_below_message.format(name, bounded_below))
        checks.append((lambda s, v: v > bounded_below, exception))

    if bounded_below_by is not None:
        exception = ValueError(bounded_below_by_message.format(name, bounded_below_by))
        checks.append(
            (lambda s, v: v > s.__getattribute__(bounded_below_by), exception)
        )

    def getter(self):
        return self.__getattribute__(hidden_name)

    def setter(self, value):

        for (value_ok, err) in checks:
            if not value_ok(self, value):
                raise err

        self.__setattr__(hidden_name, value)

    return property(fget=getter, fset=setter)


def get_parameters(f):
    """
    Given a function f, return a tuple of all its
    parameters
    """

    return tuple(signature(f).parameters.keys())


class MessageBus:
    """A class that is used behind the scenes to coordinate events and timings of
    animations.
    """

    def __init__(self):
        self.subs = {}

    def new_id(self):
        """Use this to get a name to use for your events."""
        return str(uuid4())

    def register(self, event, obj):
        """Register to receive notifications of an event.

        :param event: The name of the kind of event to receive
        :param obj: The object to receive that kind of message.
        """

        if event not in self.subs:
            self.subs[event] = [obj]
            return

        self.subs[event].append(obj)

    def send(self, event, **kwargs):
        """Send a message to whoever may be listening."""

        if event not in self.subs:
            return

        for obj in self.subs[event]:

            params = get_parameters(obj)
            values = {k: v for k, v in kwargs.items() if k in params}

            obj(**values)


_message_bus = MessageBus()


def get_message_bus():
    """A function that returns an instance of the message bus to ensure everyone uses
    the same instance."""
    return _message_bus



from stylo.domain.helpers import CartesianConversion
from stylo.domain.transform import RealDomainTransform, define_transform


class Rotation(CartesianConversion, RealDomainTransform):
    """A domain that applies a rotation to the domain provided to it."""

    def __init__(self, domain, angle):
        super().__init__(domain)
        self.angle = angle

    def _repr(self):
        return "Rotation: {}".format(self.angle)

    def _get_r(self):
        return self.domain.r

    def _get_t(self):
        ts = self.domain.t

        def mk_ts(width, height):
            return ts(width, height) + self.angle

        return mk_ts


rotate = define_transform(Rotation)



import numpy as np
from abc import ABC, abstractmethod
from textwrap import indent

from stylo.domain import RealDomain
from stylo.utils import get_parameters


class Shape(ABC):
    """A shape constructs a boolean 2D array that represents some shape or region."""

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._transforms = []

        return instance

    def __invert__(self):
        return InvertedShape(self)

    def __and__(self, other):
        return self._logical_op(ANDedShape, other)

    def __or__(self, other):
        return self._logical_op(ORedShape, other)

    def __xor__(self, other):
        return self._logical_op(XORedShape, other)

    def __call__(self, *args, **kwargs):
        return self._render(*args, **kwargs)

    def _logical_op(self, cls, other):

        if not isinstance(other, (Shape,)):
            raise TypeError("Shape: Expected Shape instance.")

        return cls(self, other)

    def _add_transform(self, transform_func):
        self._transforms.insert(0, transform_func)

    def _apply_transform(self, domain):

        if not self._transforms:
            return domain

        for transform in self._transforms:
            domain = transform(domain)

        return domain

    def _render(self, *args, **kwargs):
        """Override this function if you want to change the default evaluation rules."""

        if len(args) > 0 and isinstance(args[0], (RealDomain,)):
            return self._render_domain(*args)

        return self._render_args(**kwargs)

    def _render_domain(self, *args):

        domain = args[0]
        shape = self.draw()

        coordinates = get_parameters(shape)
        values = domain[coordinates]

        width, height = args[1], args[2]
        return shape(*values(width, height))

    def _render_args(self, **kwargs):
        shape = self.draw()
        return shape(**kwargs)

    @property
    def parameters(self):
        return get_parameters(self.draw())

    @abstractmethod
    def draw(self):
        pass


class InvertedShape(Shape):
    """A shape that has been inverted."""

    def __init__(self, shape):
        self.shape = shape

    def __repr__(self):
        return "NOT [ " + repr(self.shape) + " ]"

    def _render_domain(self, *args):
        mask = self.shape._render_domain(*args)
        return np.logical_not(mask)

    def _render_args(self, **kwargs):
        mask = self.shape._render_args(**kwargs)
        return np.logical_not(mask)

    def draw(self):
        pass

    @property
    def parameters(self):
        return self.shape.parameters


def composite_shape_factory(op, name, op_name):
    """A factory function that returns class definitions for composite shapes
    constructed using binary logical operations such as AND, OR, etc.

    :param op: The function that implements the operation in question
    :param name: The name to give the composite class.
    :param op_name: The name of the logical operation, this will be used in the object's
                    __repr__

    :type op: callable
    :type name: str
    :type op_name: str

    :rtype: class
    """

    class CompositeShape(Shape):
        """Represents a shape that has been anded together."""

        def __init__(self, a, b):
            self.a = a
            self.b = b

        def __repr__(self):
            a_repr = indent("-- " + repr(self.a), "| ")
            b_repr = indent("-- " + repr(self.b), "| ")

            return "\n".join([op_name, a_repr, b_repr])

        def _render_domain(self, *args):
            a = self.a._render_domain(*args)
            b = self.b._render_domain(*args)

            return op(a, b)

        def _render_args(self, **kwargs):

            a_params = self.a.parameters
            b_params = self.b.parameters

            a_values = {k: v for k, v in kwargs.items() if k in a_params}
            b_values = {k: v for k, v in kwargs.items() if k in b_params}

            aas = self.a._render_args(**a_values)
            bbs = self.b._render_args(**b_values)

            return op(aas, bbs)

        @property
        def parameters(self):
            a = {*self.a.parameters}
            b = {*self.b.parameters}

            return tuple(a | b)

        def draw(self):
            pass

    CompositeShape.__name__ = name
    return CompositeShape


ANDedShape = composite_shape_factory(np.logical_and, "ANDedShape", "AND")
ORedShape = composite_shape_factory(np.logical_or, "ORedShape", "OR")
XORedShape = composite_shape_factory(np.logical_xor, "XORedShape", "XOR")



"""This module contains functions that automate the process of writing base test classes
for :code:`ParameterGroup` objects.
"""
import pytest
import numpy as np
from hypothesis import given
from stylo.testing.strategies import real


PARAM_TEST_DOCSTRING = """
A base class to use when testing {0} implementations.

When writing your test case, declare it as follows:

.. code-block:: python

   from unittest import TestCase
   from stylo.testing.design import Base{0}Test

   class TestMyGroup(TestCase, Base{0}Test):
      ...

.. note::

   The order in which you include these classes is *very* important.

You also need to define a :code:`setUp()` method to set the :code:`params` attribute
to be an instance of your class.

.. code-block:: python

   def setUp(self):
       self.params = MyParams()

If you follow these steps in addition to the test cases you write your class will be
automatically checked to see that it conforms to the behavior expected from every {0}.
"""


def keys_check(expected_keys):
    """Parameter groups should define a :code:`keys()` method that returns the name of
    each of the parameters in the group.
    """

    def check(self):
        assert self.params.keys() == expected_keys

    return check


def getitem_check(keys):
    """Parameter groups should define :code:`__getitem__` and return a value for each
    parameter in the group."""

    def check(self):

        for key in keys:
            self.params[key]

    return check


def mapping_check(expected_keys):
    """Parameter groups should support the dictionary unpacking syntax :code:`**params`.
    """

    def check(self):
        def map_func(**kwargs):
            assert set(kwargs.keys()) == set(expected_keys)

        map_func(**self.params)

    return check


def parameters_check(expected_params):
    """Parameter groups should define a :code:`parameters` property that returns a list
    of the parameter names."""

    def check(self):
        assert self.params.parameters == expected_params

    return check


def parameters_check_read_only():
    """The :code:`parameters` property should be read only."""

    def check(self):
        with pytest.raises(AttributeError) as err:
            self.params.parameters = "2"

        assert "can't set attribute" in str(err.value)

    return check


def parameter_check(parameter):
    """Parameter groups should define a property for each parameter name."""

    def check(self):
        self.params.__getattribute__(parameter)

    return check


def parameter_check_read_only(parameter):
    """The declared parameter should be read only."""

    def check(self):

        with pytest.raises(AttributeError) as err:
            self.params.__setattr__(parameter, 2)

        assert "can't set attribute" in str(err.value)

    return check


def define_base_tests(pgroup):
    """Define the test cases that are common to every type of parameter group.

    :param list pgroup: The parameter group to define test cases for.
    """

    params = pgroup._parameters
    attributes = {
        "__doc__": PARAM_TEST_DOCSTRING.format(pgroup.__name__),
        "test_getitem": getitem_check(params),
        "test_keys": keys_check(params),
        "test_parameters": parameters_check(params),
        "test_parameters_read_only": parameters_check_read_only(),
        "test_mapping": mapping_check(params),
    }

    for pname in params:
        property_check = "test_{}_property".format(pname)
        read_only_check = "test_{}_property_read_only".format(pname)

        attributes[property_check] = parameter_check(pname)
        attributes[read_only_check] = parameter_check_read_only(pname)

    return attributes


def define_parameter_group_test(pgroup):
    """Given a parameter group, write the test class that will automatically verify that
    a given subclass follows the expected interface.
    """

    test_name = "Base" + pgroup.__name__ + "Test"
    attributes = define_base_tests(pgroup)

    return type(test_name, (), attributes)


def call_check(expected_keys):
    """Time dependent parameter groups should be able to be called as a function, taking
    the time :code:`t` as an argument."""

    context = given(t=real)

    def check(self, t):
        def map_check(**kwargs):
            assert set(kwargs.keys()) == set(expected_keys)

        result = self.params(t)

        # The parameter group should return itself.
        assert result is self.params
        map_check(**self.params(t))

    return context(check)


def call_check_numpy():
    """Time dependent parameter groups should be able to take numpy arrays as an
    argument."""

    ts = np.linspace(0, 1, 128)

    def check(self):
        self.params(ts)

    return check


def define_time_dependent_parameter_group_test(pgroup):
    """Given a time dependent parameter group, write the test calss that will
    automatically verify that a given subclass follows the expected interface."""

    test_name = "Base" + pgroup.__name__ + "Test"
    params = pgroup._parameters

    attributes = define_base_tests(pgroup)
    attributes["test_call"] = call_check(params)
    attributes["test_call_numpy"] = call_check_numpy()

    return type(test_name, (), attributes)



from stylo.utils import get_message_bus


class Clock:
    """This manages all the ticking etc."""

    def __init__(self, fps=25):
        self.fps = fps
        self.timedelta = 1 / fps
        self.bus = get_message_bus()

        self.time = 0
        self.tick = 0

        self.event_id = self.bus.new_id()

    def on_tick(self, obj):
        self.bus.register(self.event_id, obj)

    def __call__(self):
        self.tick += 1
        self.time += self.timedelta

        self.bus.send(self.event_id, f=self.tick, t=self.time)



"""This module defines the standard shapes you would expect to see from any graphics
package such as lines, simple curves, circles, squares, ellipses etc.
"""
import numpy as np

from stylo.shape.shape import Shape


def define_ellipse(a, b):
    """An ellipse can be implicitly defined by the equation.

    .. math::

       \\frac{x^2}{a^2} + \\frac{y^2}{b^2} = r^2

    where:

    - :math:`a` is known as the semi-major axis. Larger values stretch the ellipse along
      the :math:`x`-direction
    - :math:`b` is known as the semi-minor axis. Larger values stretch the ellipse along
      the :math:`y`-direction
    - :math:`r` is the radius and controls the overall size.

    This function will calculate the value of the left hand side of the above equation.

    :param a: The value of the semi-major axis
    :param b: The value of the semi-minor axis.

    :type a: float
    :type b: float

    :returns: The function in :code:`(x, y)` for the left hand side of the equation
    above.
    """
    a = a * a
    b = b * b

    def ellipse(x, y):
        return (x * x) / a + (y * y) / b

    return ellipse


def define_line(p1, p2):
    """Return a function than can be used to determine if a given point lies on the line
    joining the points :math:`p_1` and :math:`p_2`.

    This function assumes that your line is **not** vertical.

    The function returned will be the implicit form of a straight line

    .. math::

       f(x, y) = \\frac{(y_2 - y_1)}{(x_2 - x_1)}(x - x_1) + (y_1 - y)

    :param p1: The :math:`(x_1, y_1)` coordinate of :math:`p_1`
    :param p2: The :math:`(x_2, y_2)` coordinate of :math:`p2`

    :type p1: tuple
    :type p2: tuple
    """

    x1, y1 = p1
    x2, y2 = p2

    m = (y2 - y1) / (x2 - x1)

    def line(x, y):
        return m * (x - x1) + (y1 - y)

    return line


class Line(Shape):
    """We can define a line joining two points :math:`p_1, p_2` implicitly using the
    following equation.

    .. math::

       f(x, y) = \\frac{(y_2 - y_1)}{(x_2 - x_1)}(x - x_1) + (y_1 - y) = 0

    where:

    - :math:`x_2 \\neq x_1`
    - :math:`p_1 = (x_1, y_1)`
    - :math:`p_2 = (x_2, y_2)`

    However since a line has no area we would never see it so we need to introduce an
    error margin :math:`e > 0`. We will say a point is on the line if

    .. math::

       |f(x, y)| \\leq e
    """

    def __init__(
        self, p1=None, p2=None, pt=0.01, extend=False, above=False, below=False
    ):
        """Construct an instance of :code:`Line` depending on the the given parameters.

        By default only a line segment is drawn joining the two points :code:`p1, p2`.
        However if :code:`extend=True` then the line will be extended off to infinity.

        The thickness of the line can be controlled with the parameter :code:`pt`, which
        corresponds to :math:`e` in the equations above.

        :param p1: The first coordinate used to define the line.
                   (Default :code:`(0, 0)`)
        :param p2: The second coordinate used to define the line.
                   (Default :code:`(1, 1)`)
        :param pt: This defines the thickness of the line. (Default :code:`0.01`)
        :param extend: If :code:`True` the line will be extended off to infinity.
        :param above: If :code:`True` then the area above the line is shaded. Only takes
                      effect when :code:`extend=True`
        :param below: If code:`True` the the area below the line is shaded. Only takes
                      effect when :code:`extend=True`

        :type p1: tuple
        :type p2: tuple
        :type pt: float
        :type extend: bool
        :type above: bool
        :type below: bool
        """

        p1 = (0, 0) if p1 is None else p1
        p2 = (1, 1) if p2 is None else p2

        # It makes things easier if we can assume x1 < x2
        points = sorted([p1, p2], key=lambda p: p[0])

        self.p1 = points[0]
        self.p2 = points[1]

        self.pt = pt
        self.extend = extend
        self.above = above
        self.below = below

    def draw(self):

        pt = self.pt
        line_definition = define_line(self.p1, self.p2)

        if self.extend and self.above:

            def above_line(x, y):
                return line_definition(x, y) <= 0

            return above_line

        if self.extend and self.below:

            def below_line(x, y):
                return line_definition(x, y) >= 0

            return below_line

        if self.extend:

            def line(x, y):
                error = np.abs(line_definition(x, y))
                return error <= pt

            return line

        x1, _ = self.p1
        x2, _ = self.p2

        def line_segment(x, y):
            error = np.abs(line_definition(x, y)) <= pt

            after = x1 < x
            before = x < x2

            in_bounds = np.logical_and(after, before)
            return np.logical_and(in_bounds, error)

        return line_segment


class Ellipse(Shape):
    """One way of defining an ellipse centered at :math:`(x_0, y_0)` is through the
    implicit equation.

    .. math::

       \\frac{(x - x_0)^2}{a^2} + \\frac{(y - y_0)^2}{b^2} - r^2 = 0

    where:

    - :math:`a` is known as the semi-major axis. Larger values stretch the ellipse along
      the :math:`x`-direction
    - :math:`b` is known as the semi-minor axis. Larger values stretch the ellipse along
      the :math:`y`-direction
    - :math:`r` is the radius and controls the overall size.

    By playing around with this definition we can draw just the outline of the ellipse
    with a given thickness or a filled in version of the ellipse.

    """

    def __init__(
        self,
        x=0,
        y=0,
        a=2,
        b=1,
        r=0.5,
        pt=0.01,
        fill=False,
        start_angle=None,
        end_angle=None,
    ):
        """Construct an instance of :code:`Ellipse` based on given parameters.

        By default this will define an ellipse that only draws it's outline. If you want
        a filled in ellipse then see the :code:`fill` parameter.

        If values for :code:`start_angle` and :code:`end_angle` are provided the portion
        of the ellipse drawn will be restricted to the portion between those two angles.
        If only one of the the values are set then only an upper/lower bound will be
        imposed.

        .. note::

           All angles are in :term:`radians`.

        :param x: The :math:`x`-coordinate of the center of the ellipse. (Default:
                  :code:`0`)
        :param y: The :math:`y`-coordinate of the center of the ellipse. (Default:
                  :code:`0`)
        :param a: The length of the semi-major axis (Default: :code:`2`)
        :param b: The length of the semi-minor axis. (Default: :code:`1`)
        :param r: The radius of the ellipse. (Default: :code:`0.5`)
        :param pt: This controls the thickness of the line used to draw the ellipse.
                   (Default: :code:`0.01`). Has no effect if :code:`fill` is used.
        :param fill: If true the ellipse is drawn as a shaded region instead of a curve.
                     (Default: :code:`False`)
        :param start_angle: If set only draw the ellipse for values of
                            :math:`\\theta \\geq t` (Default: :code:`None`)
        :param end_angle: If set only draw the ellipse for values of
                          :math:`\\theta \\leq t` (Default: :code:`None`)

        :type x: float
        :type y: float
        :type a: float
        :type b: float
        :type r: float
        :type pt: float
        :type fill: bool
        :type start_angle: float
        :type end_angle: float

        :raises ValueError: If the provided arguments are inconsistent in some way.
        """

        # These define the properties of the ellipse itself.
        self.x = x
        self.y = y
        self.a = a
        self.b = b
        self.r = r

        # These define how the ellipse is actually drawn.
        self.pt = pt
        self.fill = fill

        if start_angle and end_angle and end_angle <= start_angle:
            raise ValueError("start_angle must be strictly less than end_angle")

        self.start_angle = start_angle
        self.end_angle = end_angle

    def __repr__(self):
        return "Ellipse(x={0.x}, y={0.y}, a={0.a}, b={0.b}, r={0.r})".format(self)

    def _get_filled_ellipse(self):
        """Return a function that will draw a filled in ellipse."""

        x0 = self.x
        y0 = self.y
        r = self.r * self.r
        ellipse = define_ellipse(self.a, self.b)

        def ellipse_fill(x, y):
            return ellipse(x - x0, y - y0) <= r

        return ellipse_fill

    def _get_ellipse(self):
        """Return a function that will draw an ellipse as a curve."""

        x0 = self.x
        y0 = self.y
        pt = self.pt
        r = self.r * self.r
        ellipse = define_ellipse(self.a, self.b)

        def ellipse_curve(x, y):
            error = ellipse(x - x0, y - y0) - r
            return np.abs(error) <= pt

        return ellipse_curve

    def draw(self):

        if self.fill:
            return self._get_filled_ellipse()

        return self._get_ellipse()


class Circle(Ellipse):
    """
    Mathematically a circle can be defined as the set of all
    points :math:`(x, y)` that satisfy

    .. math::

    (x - x_0)^2 + (y - y_0)^2 = r^2

    This function returns another function which when given
    a point :code:`(x, y)` will return :code:`True` if that
    point is in the circle
    """

    def __init__(self, x=0, y=0, r=0.5, *args, **kwargs):
        super().__init__(x, y, 1, 1, r, *args, **kwargs)

    def __repr__(self):
        return "Circle(x={0.x},y={0.y},r={0.r})".format(self)


class Rectangle(Shape):
    """
    It's quite simple to define a rectangle, simply pick a
    point :math:`(x_0,y_0)` that you want to be the center
    and then two numbers which will represent the width and
    height of the rectangle.
    """

    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __repr__(self):
        arg_string = "x={0.x},y={0.y},width={0.width},height={0.height}".format(self)
        return "Rectangle({})".format(arg_string)

    def draw(self):

        left = self.x - (self.width / 2)
        right = self.x + (self.width / 2)
        top = self.y + (self.height / 2)
        bottom = self.y - (self.height / 2)

        def rectangle(x, y):
            xs = np.logical_and(left < x, x < right)
            ys = np.logical_and(bottom < y, y < top)

            return np.logical_and(xs, ys)

        return rectangle


class Square(Rectangle):
    def __init__(self, x, y, size):
        super().__init__(x, y, size, size)

    def __repr__(self):
        return "Square(x={0.x},y={0.y},size={0.width})".format(self)


class Triangle(Shape):
    """
    A Triangle can be defined by picking three non-collinear points
    :math:`(a, b, c)` in the form of a tuple each.
    Points inside the triangle are determined using the
    barycentric coordinate system method, which is further explained
    `in this answer <https://math.stackexchange.com/a/1884485/294670>`_
    and the first method
    `here <http://totologic.blogspot.com/2014/01/accurate-point-in-triangle-test.html>`_.
    """  # noqa E501

    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

    def __repr__(self):
        return "Triangle(a={0.a}, b={0.b}, c={0.c})".format(self)

    def get_q(self):
        """
        Returns the denominator for calculating both the barycentric
        coordinates, which absolute value is the double of the area
        of the triangle.
        """
        a = self.a
        b = self.b
        c = self.c
        q = -b[1] * c[0] + a[1] * (-b[0] + c[0]) + a[0] * (b[1] - c[1]) + b[0] * c[1]

        return q

    def get_s(self, x, y):
        """
        Returns one of the barycentric coordinates of the triangle using both
        :math:`a` and :math:`c` points.
        """
        a = self.a
        c = self.c
        sign = -1 if self.get_q() < 0 else 1
        s = (a[1] * c[0] - a[0] * c[1] + (c[1] - a[1]) * x + (a[0] - c[0]) * y) * sign
        return s

    def get_t(self, x, y):
        """
        Returns one of the barycentric coordinates of the triangle using both
        :math:`a` and :math:`b` points.
        """
        a = self.a
        b = self.b
        sign = -1 if self.get_q() < 0 else 1
        t = (a[0] * b[1] - a[1] * b[0] + (a[1] - b[1]) * x + (b[0] - a[0]) * y) * sign
        return t

    def draw(self):
        def triangle(x, y):
            sign = -1 if self.get_q() < 0 else 1
            first_condition = self.get_s(x, y) > 0
            second_condition = self.get_t(x, y) > 0
            third_condition = self.get_s(x, y) + self.get_t(x, y) < self.get_q() * sign

            return np.logical_and(
                first_condition, np.logical_and(second_condition, third_condition)
            )

        return triangle



from unittest import TestCase

from stylo.color import FillColor
from stylo.testing.color import BaseColorMapTest


class TestFillColor(TestCase, BaseColorMapTest):
    """Tests for the :code:`FillColor` colormap."""

    def setUp(self):
        self.colormap = FillColor()

    def test_init_no_args(self):
        """Ensure that if no arguments are given, then the color defaults to black."""

        fill = FillColor()
        self.assertEqual((0, 0, 0), fill.color)



import pytest
from hypothesis import given, assume

from stylo.math import lerp
from stylo.testing.strategies import real


@pytest.mark.math
class TestLerp:
    """Tests for the :code:`lerp` function."""

    @given(a=real, b=real)
    def test_take_a_at_zero(self, a, b):
        """Ensure that the resulting function takes the value :code:`a` at :code:`0`."""

        assume(a != b)

        s = lerp(a, b)
        assert s(0) == a

    @given(a=real, b=real)
    def test_take_b_at_one(self, a, b):
        """Ensure that the resulting function takes the value :code:`b` at :code:`1`."""

        assume(a != b)

        s = lerp(a, b)
        assert s(1) == b



import numpy as np

from stylo.domain import RealDomain
from stylo.domain.helpers import PolarConversion
from stylo.utils import bounded_property


class RectangularDomain(PolarConversion, RealDomain):
    """A domain of the form :math:`[a, b] \\times [c, d] \\subset \\mathbb{R}^2`"""

    xmin = bounded_property("xmin", bounded_above_by="xmax")
    xmax = bounded_property("xmax", bounded_below_by="xmin")
    ymin = bounded_property("ymin", bounded_above_by="ymax")
    ymax = bounded_property("ymax", bounded_below_by="ymin")

    def __init__(self, xmin, xmax, ymin, ymax):

        if xmin >= xmax:
            raise ValueError(
                "The value of xmin must be strictly less than the " "value of xmax"
            )

        if ymin >= ymax:
            raise ValueError(
                "The value of ymin must be strictly less than the " "value of ymax"
            )

        self._xmin = xmin
        self._ymin = ymin
        self._xmax = xmax
        self._ymax = ymax

    def __repr__(self):
        name = self.__class__.__name__
        return "{1}: [{0.xmin}, {0.xmax}] x [{0.ymin}, {0.ymax}]".format(self, name)

    def _get_x(self):
        def mk_xs(width, height):
            xs = np.linspace(self.xmin, self.xmax, width)
            xs = np.array([xs for _ in range(height)])

            return xs

        return mk_xs

    def _get_y(self):
        def mk_ys(width, height):
            ys = np.linspace(self.ymax, self.ymin, height)
            ys = np.array([ys for _ in range(width)])
            ys = ys.transpose()

            return ys

        return mk_ys


def get_real_domain(width, height, scale=2):
    """Given the width and height for an image return a rectangular domain with
    appropriate bounds for the image's aspect ratio.

    So that shapes are not distorted when mapped onto non-square images the domain they
    are mapped onto needs to match the aspect ratio of the image otherwise the shape
    will be distorted in some way.

    This function will construct a RectangularDomain domain with :math:`(0, 0)` at the
    centre of the image, the overall size of the domain can be controlled with the
    :code:`scale` parameter

    :param width: The width of the image in pixels
    :param height: The height of the image in pixels
    :param scale: This constrols the size of the domain and corresponds to the length of
                  the interval :math:`[ymin, ymax]`

    :type width: int
    :type height: int
    :type scale: float

    :returns: An appropriately sized domain.
    :rtype: RectangularDomain
    """

    if height <= 0:
        raise ValueError("The height of the image must be a positive number.")

    if width <= 0:
        raise ValueError("The width of the image must be a positive number.")

    if scale <= 0:
        raise ValueError("The scale of the image must be strictly positive.")

    aspect_ratio = width / height

    ylength = scale / 2
    xlength = ylength * aspect_ratio

    ymin, ymax = -ylength, ylength
    xmin, xmax = -xlength, xlength

    return RectangularDomain(xmin, xmax, ymin, ymax)



import pytest
from unittest import TestCase
from hypothesis import given

from stylo.design import StaticPosition, ParametricTrajectory
from stylo.testing.design import BasePositionTest, BaseTrajectoryTest
from stylo.testing.strategies import real


@pytest.mark.pgroup
class TestStaticPosition(TestCase, BasePositionTest):
    """Tests for the :code:`StaticPosition` parameter group."""

    def setUp(self):
        self.params = StaticPosition(x=0.4, y=-0.8)

    def test_default_init(self):
        """If no values are given then :code:`StaticPosition` should choose sensible
        defaults."""

        params = StaticPosition()

        def f(**kwargs):
            assert kwargs == {"x": 0, "y": 0}

        f(**params)

    @given(x=real, y=real)
    def test_init(self, x, y):
        """:code:`StaticPosition` should use the values supplied."""

        params = StaticPosition(x, y)

        def f(**kwargs):
            assert kwargs == {"x": x, "y": y}

        f(**params)


@pytest.mark.pgroup
class TestParametricTrajectory(TestCase, BaseTrajectoryTest):
    """Tests for the :code:`ParametricTrajectory` parameter group."""

    def setUp(self):
        def x(t):
            return 2 * t

        def y(t):
            return t * t - 4 * t + 3

        self.params = ParametricTrajectory(x, y)

    @given(t=real)
    def test_default_init(self, t):
        """If no functions are given then :code:`ParametricTrajectory` should choose
        sensible defaults."""

        params = ParametricTrajectory()

        def f(**kwargs):
            assert kwargs == {"x": t, "y": t}

        f(**params(t))

    @given(t=real)
    def test_init(self, t):
        """:code:`ParametricTrajectory` should use the functions supplied."""

        def x(s):
            return 2 * s

        def y(s):
            return s / 2

        params = ParametricTrajectory(x, y)

        def f(**kwargs):
            assert kwargs == {"x": 2 * t, "y": t / 2}

        f(**params(t))



from stylo.domain.helpers import PolarConversion
from stylo.domain.transform import RealDomainTransform, define_transform


class HorizontalShear(PolarConversion, RealDomainTransform):
    """A horizontal shear transform."""

    def __init__(self, domain, k=0):
        super().__init__(domain)
        self.k = -k

    def _repr(self):
        return "HorizontalShear: {}".format(-self.k)

    def _get_y(self):
        return self.domain.y

    def _get_x(self):

        xs = self.domain.x
        ys = self.domain.y

        def mk_xs(width, height):
            return xs(width, height) + self.k * ys(width, height)

        return mk_xs


class VerticalShear(PolarConversion, RealDomainTransform):
    """A vertical shear transform."""

    def __init__(self, domain, k=0):
        super().__init__(domain)
        self.k = -k

    def _repr(self):
        return "VerticalShear: {}".format(-self.k)

    def _get_x(self):
        return self.domain.x

    def _get_y(self):

        xs = self.domain.x
        ys = self.domain.y

        def mk_ys(width, height):
            return ys(width, height) + self.k * xs(width, height)

        return mk_ys


vertical_shear = define_transform(VerticalShear)
horizontal_shear = define_transform(HorizontalShear)



import pytest
import numpy.testing as npt
from hypothesis import given
from unittest import TestCase

from stylo.domain import UnitSquare
from stylo.domain.transform.shear import HorizontalShear, VerticalShear
from stylo.testing.domain import BaseRealDomainTest
from stylo.testing.strategies import real, dimension


@pytest.mark.domain
class TestHorizontalShear(TestCase, BaseRealDomainTest):
    """Tests for the horizontal shear class."""

    def setUp(self):
        self.domain = HorizontalShear(UnitSquare(), 0.5)

    @given(width=dimension, height=dimension, k=real)
    def test_k_affects_x_values(self, width, height, k):
        """Ensure that :code:`k` shears the x values by the factor -ky so that the
        user's intuition is maintained. Also ensure that the y values are unaffected."""

        domain = UnitSquare()
        xs, ys = domain["xy"](width, height)

        sheared = HorizontalShear(domain, k)
        t_xs, t_ys = sheared["xy"](width, height)

        # Y values should be unchanged
        npt.assert_array_equal(ys, t_ys)

        # X values should only differ by -ky
        difference = t_xs - xs
        npt.assert_almost_equal(difference, -k * ys)


@pytest.mark.domain
class TestVerticalShear(TestCase, BaseRealDomainTest):
    """Tests for the vertical shear class."""

    def setUp(self):
        self.domain = VerticalShear(UnitSquare(), 0.5)

    @given(width=dimension, height=dimension, k=real)
    def test_k_affects_y_values(self, width, height, k):
        """Ensure that :code:`k` shears the y values by a factor of -kx so that the
        user's intuition is maintained. Also ensure that the x values are unaffected."""

        domain = UnitSquare()
        xs, ys = domain["xy"](width, height)

        sheared = VerticalShear(domain, k)
        t_xs, t_ys = sheared["xy"](width, height)

        # X values should be unchanged
        npt.assert_array_equal(xs, t_xs)

        # Y values should only differ by -kx
        difference = t_ys - ys
        npt.assert_almost_equal(difference, -k * xs)



"""A :code:`ParameterGroup` is used to help define and manage a related set of
parameters. For example a "Position" parameter group could contain values for the
:code:`x` and :code:`y` position of an object. Another "Stickman" property group could
contain all the values required to define a stickman with a particular size and shape.

This module much like the :code:`stylo.domain._factory` module defines a number of
functions for declaring new parameter groups of various types. By writing the base
classes using code, we can ensure that all property groups follow the same interface.

Currently the following functions are defined

- :code:`define_parameter_group`: This defines a base, static parameter group.
- :code:`define_time_dependent_parameter_group`: This defines a time-dependent parameter
  group, that can take some time :code:`t` and vary in time. It also provides some
  utilities for visualising these changes over time.
"""
from abc import ABC, abstractmethod
import numpy as np
import matplotlib.pyplot as plt


def pgroup_keys(params):
    """Provide the implementation of the :code:`keys` method for this parameter group

    One of the key features of a parameter group is the ability to use the dictionary
    unpacking syntax (:code:`**params`) to conveniently pass all the values into a
    function as a single unit. This syntax is enabled by implementing two methods
    :code:`keys` and :code:`__getitem__`, this function provides the implementation
    of the former.

    :param list params: The list of strings containing the names of the parameters.
    """

    def keys(self):
        return params

    return keys


def pgroup_parameters_prop():
    """Provide the implementation of the :code:`parameters` property.

    The :code:`parameters` property returns a list of all the parameter names in the
    group.
    """

    @property
    def parameters(self):
        return self.keys()

    return parameters


def pgroup_getitem():
    """Provide the implementation of the :code:`__getitem__` method.

    One of the key features of a parameter group is the ability to use the dictionary
    unpacking syntax (:code:`**params`) to conveniently pass all the values into a
    function as a single unit. This syntax is enabled by implementing two methods
    :code:`keys` and :code:`__getitem__`, this function provides the implementation
    of the latter.
    """

    def __getitem__(self, key):
        return self._values[key]

    return __getitem__


def pgroup_init():
    """Provide the implementation of the :code:`__init__` method."""

    def __init__(self):
        self._values = self.calculate()

    return __init__


def pgroup_parameter_prop(param):
    """Provide the implementation of the property getter for the given parameter name.

    This enables the :code:`group.param` syntax.

    :param str param: The parameter name.
    """

    @property
    def parameter(self):
        return self._values[param]

    return parameter


def pgroup_calculate():
    """Define an abstract method for the user to implement.

    The user must define this method to return the values dictionary containing each
    of the parameters in the group,.
    """

    @abstractmethod
    def calculate(self):
        pass

    return calculate


def pgroup_repr(name, params):
    """Provide the implementation of the :code:`__repr__` method.

    :param str name: The name of the parameter group.
    :param: list params: The list of parameter names.
    """

    offset = max([len(p) for p in params]) + 10

    def __repr__(self):

        cls_name = self.__class__.__name__

        values = [
            p + ":" + str(val).rjust(offset - len(p)) for p, val in self._values.items()
        ]

        ps = "\n  ".join(values)
        return "{}: {}\n  {}".format(cls_name, name, ps)

    return __repr__


def define_base_attributes(name, params):
    """Define the attributes, methods and properties that are common to every type of
    parameter group.

    :param str name: The name of the parameter group
    :param list params: The list of strings containing the parameter names.
    """

    attributes = {
        "__doc__": "TODO: Make this docstring useful.",
        "__getitem__": pgroup_getitem(),
        "__repr__": pgroup_repr(name, params),
        "keys": pgroup_keys(params),
        "parameters": pgroup_parameters_prop(),
        "_parameters": params,
    }

    for p in params:
        attributes[p] = pgroup_parameter_prop(p)

    return attributes


def get_parameter_names(parameters):
    """Convert the parameter string supplied by the user into a list of parameter names.

    :param list parameters: The comma separated string of parameter names
    """
    return parameters.replace(" ", "").split(",")


def define_parameter_group(name, parameters):
    """Define a new parameter group.

    :param str name: The name of the parameter group.
    :param list parameters: A list of strings containing the parameter
                            names for the group.
    """

    params = get_parameter_names(parameters)

    attributes = define_base_attributes(name, params)
    attributes["__init__"] = pgroup_init()
    attributes["calculate"] = pgroup_calculate()

    return type(name, (ABC,), attributes)


def tdpgroup_calculate():
    """Define an abstract :code:`calculate` method for users to implement.

    This method will serve the same role as the calculate method in a standard parameter
    group, but this one is dependent on time :code:`t`.
    """

    @abstractmethod
    def calculate(self, t):
        pass

    return calculate


def tdpgroup_init():
    """Provide the implementation of the :code:`__init__` method."""

    def __init__(self):
        self()

    return __init__


def tdpgroup_call():
    """Provide the implementation of the :code:`__call__` method.

    This enables the :code:`**params(t)` functionality and use the parameter group as a
    time dependent function.
    """

    def __call__(self, t=0):
        self._values = self.calculate(t)
        return self

    return __call__


def tdpgroup_plot():
    """Provide the implementation of the :code:`plot` method.

    This enables users to plot the values as they change over time.
    """

    def plot(self, start=0, stop=1, N=128, plot_size=8):

        cls = self.__class__
        name = cls.__name__
        base_name = cls.__base__.__name__

        ts = np.linspace(start, stop, N)
        values = self.calculate(ts)

        fig, ax = plt.subplots(1, figsize=(plot_size, plot_size))

        for pname, vals in values.items():
            ax.plot(ts, vals, label=pname)

        # Include a legend, title and label the time axis.
        ax.legend()
        ax.set_title("{}: {}".format(name, base_name))
        ax.set_xlabel("Time (seconds)")

        return ax

    return plot


def define_time_dependent_parameter_group(name, parameters):
    """Define a new time dependent parameter group.

    :param str name: The name of the parameter group
    :param list parameters: A list of strings containing the parameter names for the
                            group.
    """
    params = get_parameter_names(parameters)
    attributes = define_base_attributes(name, params)

    attributes["__call__"] = tdpgroup_call()
    attributes["__init__"] = tdpgroup_init()
    attributes["calculate"] = tdpgroup_calculate()
    attributes["plot"] = tdpgroup_plot()

    return type(name, (ABC,), attributes)



import pytest
from hypothesis import given, assume
from unittest import TestCase

from stylo.testing.strategies import real
from stylo.utils import bounded_property


@pytest.mark.utils
class TestBoundedProperty(TestCase):
    """Tests to ensure the :code:`bounded_property` factory function produces
    properties as expected."""

    def test_checks_bounded_below_type(self):
        """Ensure that :code:`bounded_property` informs the user when using a value
        for :code:`bounded_below` with the incorrect type."""

        with pytest.raises(TypeError) as err:
            bounded_property("length", bounded_below="width")

        self.assertIn("must be a number", str(err.value))
        self.assertIn("bounded_below", str(err.value))

    def test_checks_bounded_below_by_type(self):
        """Ensure that :code:`bounded_property` informs the user when using
        a value for :code:`bounded_below_by` with the incorrect type."""

        with pytest.raises(TypeError) as err:
            bounded_property("length", bounded_below_by=2.0)

        self.assertIn("must be a string", str(err.value))
        self.assertIn("bounded_below_by", str(err.value))

    def test_checks_bounded_above_type(self):
        """Ensure that :code:`bounded_property` informs the user when using
        a value for :code:`bounded_above` with the incorrect type."""

        with pytest.raises(TypeError) as err:
            bounded_property("length", bounded_above="width")

        self.assertIn("must be a number", str(err.value))
        self.assertIn("bounded_above", str(err.value))

    def test_checks_bounded_above_by_type(self):
        """Ensure that :code:`bounded_property` informs the user when
        using a value for :code:`bounded_above_` with the incorrect type."""

        with pytest.raises(TypeError) as err:
            bounded_property("length", bounded_above_by=2.0)

        self.assertIn("must be a string", str(err.value))
        self.assertIn("bounded_above_by", str(err.value))

    def test_checks_above_arguments(self):
        """Ensure that :code:`bounded_property` informs the user that it only
        makes sense to use either :code:`bounded_above_by` or :code:`bounded_above`"""

        with pytest.raises(ValueError) as err:
            bounded_property("name", bounded_above=1, bounded_above_by="prop")

        self.assertIn("You can only use", str(err.value))

    def test_checks_below_arguments(self):
        """Ensure that :code:`bounded_below` informs the user that it only
        makes sense to use either :code:`bounded_below` or :code:`bounded_below_by`"""

        with pytest.raises(ValueError) as err:
            bounded_property("name", bounded_below=1, bounded_below_by="prop")

        self.assertIn("You can only use", str(err.value))

    @given(value=real)
    def test_uses_getter(self, value):
        """Ensure that the property constructed by :code:`bounded_property`
        returns the value it manages"""

        class MyClass:
            _length = value
            length = bounded_property("length")

        my_class = MyClass()
        self.assertEqual(value, my_class.length)

    @given(value=real)
    def test_uses_setter(self, value):
        """Ensure that the property constructed by :code:`bounded_property`
        still allows the user to set the value even if no bounds are specified."""

        class MyClass:
            length = bounded_property("length")

        my_class = MyClass()
        my_class.length = value

        self.assertEqual(value, my_class.length)

    def test_setter_checks_type(self):
        """Ensure that the property constructed by :code:`bounded_property`
        checks the type of the value provided."""

        class MyClass:
            length = bounded_property("length")

        my_class = MyClass()

        with pytest.raises(TypeError) as err:
            my_class.length = "10m"

        self.assertIn("length", str(err.value))
        self.assertIn("must be a number", str(err.value))

    @given(value=real, bound=real)
    def test_setter_checks_bounded_below(self, value, bound):
        """Ensure that when given a lower bound via :code:`bounded_below` that
        the property constructed by :code:`bounded_property` checks the value
        against that bound."""

        assume(value <= bound)

        class MyClass:
            length = bounded_property("length", bounded_below=bound)

        my_class = MyClass()

        with pytest.raises(ValueError) as err:
            my_class.length = value

        self.assertIn("must be strictly larger than", str(err.value))
        self.assertIn(str(bound), str(err.value))

        value = 1 - value
        bound = -bound

        # Now we have value > bound
        class MyOtherClass:
            length = bounded_property("length", bounded_below=bound)

        my_other_class = MyOtherClass()
        my_other_class.length = value

        self.assertEqual(value, my_other_class.length)

    @given(value=real, bound=real)
    def test_setter_checks_bounded_below_by(self, value, bound):
        """Ensure that when given an attribute to act as a lower bound via
        :code:`bounded_below_by` that the property constructed by
        :code:`bounded_property` checks the value against that bound."""

        assume(value <= bound)

        class MyClass:
            width = bound
            length = bounded_property("length", bounded_below_by="width")

        my_class = MyClass()

        with pytest.raises(ValueError) as err:
            my_class.length = value

        self.assertIn("must be strictly larger than", str(err.value))
        self.assertIn("width", str(err.value))

        value = 1 - value
        bound = -bound

        # Now we have value > bound
        class MyOtherClass:
            width = bound
            length = bounded_property("length", bounded_below_by="width")

        my_other_class = MyOtherClass()
        my_other_class.length = value

        self.assertEqual(value, my_other_class.length)

    @given(value=real, bound=real)
    def test_setter_checks_bounded_above(self, value, bound):
        """Ensure that when given an upper bound via :code:`bounded_above`
        that the property constructed by :code:`bounded_property` checks the
        value against that bound."""

        assume(value >= bound)

        class MyClass:
            length = bounded_property("length", bounded_above=bound)

        my_class = MyClass()

        with pytest.raises(ValueError) as err:
            my_class.length = value

        self.assertIn("must be strictly less than", str(err.value))
        self.assertIn(str(bound), str(err.value))

        value = -1 - value
        bound = -bound

        # We now have value < bound
        class MyOtherClass:
            length = bounded_property("length", bounded_above=bound)

        my_other_class = MyOtherClass()
        my_other_class.length = value

        self.assertEqual(value, my_other_class.length)

    @given(value=real, bound=real)
    def test_setter_checks_bounded_above_by(self, value, bound):
        """Ensure that when given an upper bound via :code:`bounded_above_by`
        that the property constructed by :code:`bounded_property` checks the
        value against the bound.
        """

        assume(value >= bound)

        class MyClass:
            width = bound
            length = bounded_property("length", bounded_above_by="width")

        my_class = MyClass()

        with pytest.raises(ValueError) as err:
            my_class.length = value

        self.assertIn("must be strictly less than", str(err.value))
        self.assertIn("width", str(err.value))

        value = -1 - value
        bound = -bound

        # Now we have value < bound
        class MyOtherClass:
            width = bound
            length = bounded_property("length", bounded_above_by="width")

        my_other_class = MyOtherClass()
        my_other_class.length = value

        self.assertEqual(value, my_other_class.length)



from stylo.domain.helpers import PolarConversion
from stylo.domain.transform import RealDomainTransform, define_transform


class Translation(PolarConversion, RealDomainTransform):
    """A domain that applies a translation to the domain provided to it."""

    def __init__(self, domain, dx, dy):
        super().__init__(domain)
        self.dx = -dx
        self.dy = -dy

    def _repr(self):
        return "Translation: ({0}, {1})".format(-self.dx, -self.dy)

    def _get_x(self):

        xs = self.domain.x

        def mk_xs(width, height):
            return xs(width, height) + self.dx

        return mk_xs

    def _get_y(self):

        ys = self.domain.y

        def mk_ys(width, height):
            return ys(width, height) + self.dy

        return mk_ys


translate = define_transform(Translation)



"""A :code:`Domain` is responsible for converting some abstract notion of a mathematical
space into a discrete representation. Since :code:`stylo` is about creating images this
representation is typically a 2D grid of numbers.

This document details the machinery behind the :code:`Domain` system. More specifically
how the :code:`define_domain` and :code:`define_domain_transform` functions work. It
assumes you are already familiar with how the user interacts with :code:`Domain` objects
and is written for someone who wishes to understand how they work behind the scenes.

This module contains all the code that is required to automatically write both the
definition of a new Domain type and the test case that users can use to verify that
their implementations of said domains adhere to the expected interface.
"""
from abc import ABC, abstractmethod
from textwrap import indent


def domain_getitem(self, key):
    """Enable the domain[...] syntax for domains.

    Domains support the ability to return a function in :code:`(width, height)`
    of multiple parameters, for example for a :code:`RealDomain` you might want a
    function in both the :math:`x` and :math:`y` coordinates in which case you would
    write

    .. code-block:: python

       >>> values = domain['xy']
       >>> values(N, N)
       (array([[x0, x1, ..., xN],
               [x0, x1, ..., XN],
               ...
               [x0, x1, ..., xN]]),
        array([[y0, y0, ..., y0],
               [y1, y1, ..., y1],
               ...,
               [yN, yN, ..., yN]]))

    .. note::

       The string argument :code:`'xy'` only works in the case of single letter
       parameters. For multi-character parameters such as :code:`pt` this argument
       must be some other iterable such as a tuple e.g. :code:`('x', 'pt')`.

    This function constructs the implementation for the :code:`__getitem__` method that
    will be added to new :code:`Domain` definitions

    :param self: The standard python instance argument for methods
    :param key: The input that is given to the getitem syntax e.g. :code:`'xy'`

    :type self: object
    :type key: iterable
    """

    return lambda w, h: tuple(
        self.__getattribute__(p)(w, h) for p in key if p in self._parameters
    )


def domain_property(base):
    """Construct the domain property for the domain transform base class.

    :param base: The base Domain class
    :type base: class
    """

    transform_name = base.__name__ + "Transform"

    def getter(self):
        return self._domain

    def setter(self, value):

        if not isinstance(value, (base,)):
            message = "{transform}: Expected {base} instance."
            raise TypeError(
                message.format(transform=transform_name, base=base.__name__)
            )

        self._domain = value

    return property(fget=getter, fset=setter)


def domain_call(self, width, height):
    return lambda parameters: self[parameters](width, height)


def parameters_property(parameters):
    """Given the list of parameters this function constructs a property that simply
    returns the given list. It doesn't provide a setter so that the list of parameters
    cannot be overridden."""

    def getter(self):
        return parameters

    return property(fget=getter)


def parameter_property(name):
    """Given the name of the parameter, construct the property definition that calls the
    abstract :code:`_get_name()` method that users will implement.

    :param name: The name of the parameter
    :type name: str
    """

    def getter(self):
        return self.__getattribute__("_get_" + name)()

    return property(fget=getter)


def mk_abstractmethod():
    """Construct an abstractmethod that user must override when they create an instance
    of a class.
    """

    def method(self):
        pass

    return abstractmethod(method)


def transform_init(self, domain):
    """This implements the :code:`__init__` method for DomainTransform classes."""
    self.domain = domain


def transform_repr(self):
    """This implements the :code:`__repr__` method for DomainTransform classes."""
    transform = self._repr()
    domain = repr(self.domain)

    return transform + "\n" + indent(domain, "  ")


def define_domain(name, parameters):
    """Given the name and the input parameters, construct the Domain definition.

    :param name: The name of the :code:`Domain`
    :param parameters: A list of strings that encode the names of the parameters within
           the domain

    :type name: str
    :type parameters: list(str)
    """
    params = parameters.split(",")

    attributes = {
        "_parameters": params,
        "parameters": parameters_property(params),
        "__doc__": "TODO: Make this docstring useful.",
        "__getitem__": domain_getitem,
        "__call__": domain_call,
    }

    for p in params:
        attributes[p] = parameter_property(p)
        attributes["_get_" + p] = mk_abstractmethod()

    return type(name, (ABC,), attributes)


def define_domain_transform(base):
    """This constructs the class definition for a domain transform.

    :param name: The name of the base domain
    :param base: The class definition of the base domain

    :type name: str
    :type base: class
    """

    transform_name = base.__name__ + "Transform"

    attributes = {
        "__init__": transform_init,
        "__repr__": transform_repr,
        "_repr": mk_abstractmethod(),
        "domain": domain_property(base),
    }

    return type(transform_name, (base, ABC), attributes)



import pytest
import numpy as np
from unittest import TestCase

from stylo.domain import SquareDomain
from stylo.shape.curves import ImplicitXY
from stylo.testing.shape import BaseShapeTest


@pytest.mark.shape
class TestImplicitXY(TestCase, BaseShapeTest):
    """Tests for the :code:`ImplicitXY` shape."""

    def setUp(self):
        self.shape = ImplicitXY(lambda x, y: x - y)

    def test_init_defaults(self):
        """Ensure that the shape is initialised with sane defaults."""

        curve = ImplicitXY(lambda x, y: x * y)

        assert curve.pt == 0.01
        assert not curve.above
        assert not curve.below

    def test_draw_pt(self):
        """Ensure that we can draw a function as a curve."""

        domain = SquareDomain(-1, 1)
        curve = ImplicitXY(lambda x, y: x - y, pt=0.05)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False,  True],
                [False, False, False, False, False, False,  True, False],
                [False, False, False, False, False,  True, False, False],
                [False, False, False, False,  True, False, False, False],
                [False, False, False,  True, False, False, False, False],
                [False, False,  True, False, False, False, False, False],
                [False,  True, False, False, False, False, False, False],
                [ True, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on

        assert (expected == curve(domain, 8, 8)).all()

    def test_draw_above(self):
        """Ensure that we can draw the area above the curve as a shaded region."""

        domain = SquareDomain(-1, 1)
        curve = ImplicitXY(lambda x, y: x - y, above=True)

        # fmt: off
        expected = np.array(
            [
                [True,  True,  True,  True,  True,  True,  True,  True],
                [True,  True,  True,  True,  True,  True,  True, False],
                [True,  True,  True,  True,  True,  True, False, False],
                [True,  True,  True,  True,  True, False, False, False],
                [True,  True,  True,  True, False, False, False, False],
                [True,  True,  True, False, False, False, False, False],
                [True,  True, False, False, False, False, False, False],
                [True, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on

        assert (expected == curve(domain, 8, 8)).all()

    def test_draw_below(self):
        """Ensure that we can draw the are below the curve as a shaded region."""

        domain = SquareDomain(-1, 1)
        curve = ImplicitXY(lambda x, y: x - y, below=True)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False,  True],
                [False, False, False, False, False, False, False,  True],
                [False, False, False, False, False, False,  True,  True],
                [False, False, False, False, False,  True,  True,  True],
                [False, False, False, False,  True,  True,  True,  True],
                [False, False, False,  True,  True,  True,  True,  True],
                [False, False,  True,  True,  True,  True,  True,  True],
                [ True,  True,  True,  True,  True,  True,  True,  True],
            ]
        )
        # fmt: on

        assert (expected == curve(domain, 8, 8)).all()



import pytest
from unittest import TestCase

from stylo.shape import Circle
from stylo.color import FillColor
from stylo.image import SimpleImage
from stylo.testing.image import BaseImageTest


@pytest.mark.image
class TestSimpleImage(TestCase, BaseImageTest):
    """Tests for the :code:`SimpleImage` object."""

    def setUp(self):
        self.image = SimpleImage(Circle(0, 0, 0.9), FillColor())



import numpy as np
from hypothesis import given

from stylo.testing.strategies import shape_mask


class BaseColorMapTest:
    """A base class for testing :code:`ColorMap` implementations.

    When writing your test case for a new :code:`Shape` implementation you need to
    declare it as follows.

    .. code-block:: python

       from unittest import TestCase
       from stylo.testing.color import BaseColorMapTest

       class TestMyColorMap(TestCase, BaseColorMapTest):
           ...

    .. note::

       The order in which you write the classes is *very* important.

    You also need to define a :code:`setUp` method to set the :code:`colormap` attribute
    equal to an instance of your shape implementation

    .. code-block:: python

       def setUp(self):
           self.colormap = MyColorMap()

    Then in addition to any tests your write, your :code:`ColorMap` implementation will
    be automatically tested to see if it passes the checks defined below.
    """

    @given(mask=shape_mask)
    def test_paint(self, mask):
        """Ensure that a colormap can be called with a mask produced by some shape and
        that the result is:

        - A numpy array with the same shape :code:`(height, width)` as the given mask.

        .. note::

           Since :code:`ColorMaps` need to be independent of :code:`ColorSpace` we
           cannot enforce anything about the contents of the array
        """
        colormap = self.colormap
        colorspace = colormap.colorspace
        background = colorspace.parse("ffffff")

        height, width = mask.shape
        dimensions = (height, width, len(background))

        color = np.full(dimensions, background)
        color = colormap(mask, image_data=color)

        self.assertEqual(mask.shape[0], color.shape[0])
        self.assertEqual(mask.shape[1], color.shape[1])



import pytest

from stylo.domain import RealDomain, UnitSquare
from stylo.domain.transform import RealDomainTransform
from stylo.domain.transform.translation import Translation
from stylo.domain.transform.transform import (
    find_base_transform,
    find_base_domain,
    DomainTransformer,
)
from stylo.shape import Circle


@pytest.mark.domain
class TestFindBaseTransform:
    """Tests for the :code:`find_base_transform` function."""

    def test_with_transform(self):
        """Ensure that when given a domain transform, this function returns the
        appropriate base class."""

        base_transform = find_base_transform(Translation)
        assert base_transform == RealDomainTransform

    def test_with_shape(self):
        """Ensure that when given something that is not a domain transform, this
        function raises the appropriate exception"""

        with pytest.raises(TypeError) as err:
            find_base_transform(Circle)

        assert "is not a domain transform" in str(err.value)


@pytest.mark.domain
class TestFindBaseDomain:
    """Tests for the :code:`find_base_domain` function."""

    def test_with_base_transform(self):
        """Ensure that when given a base domain transform, this function performs as
        expected."""

        base_domain = find_base_domain(RealDomainTransform)
        assert base_domain == RealDomain

    def test_with_shape(self):
        """Ensure that when given something that is not a domain, this function raises
        the appropriate exception."""

        with pytest.raises(TypeError) as err:
            find_base_domain(Circle)

        assert "is not a base domain transform" in str(err.value)

    def test_with_transform(self):
        """Ensure that when given a transform that is not the base this raises the
        appropriate exception."""

        with pytest.raises(TypeError) as err:
            find_base_domain(Translation)

        assert "is not a base domain transform" in str(err.value)


@pytest.mark.domain
class TestDomainTransformer:
    """Tests for the :code:`DomainTransformer` class."""

    def test_apply_transform_bad_type(self):
        """Ensure that an exception is raised if the transform is applied to an
        unsupported type."""

        transformer = DomainTransformer(Translation, 1, 1)

        with pytest.raises(TypeError, match="Unable to perform"):
            transformer.apply_transform(3)

    def test_apply_transform_with_domain(self):
        """Ensure that a transform is properly applied to a domain object."""

        domain = UnitSquare()
        transform = DomainTransformer(Translation, 1, 2)

        transformed = transform.apply_transform(domain)

        assert -1 == transformed.dx
        assert -2 == transformed.dy
        assert domain == transformed.domain



"""Specific hypothesis strategies that are useful for testing.

It defines the following strategies:

- :code:`dimension`: Represents a size or dimension e.g. for numpy arrays image
  sizes etc.
- :code:`real`: Represents a real number in the range +/-1 million
"""
from math import pi

import numpy as np
from hypothesis.strategies import integers, floats, tuples
from hypothesis.extra.numpy import arrays

# Basic types
real = floats(min_value=-1e6, max_value=1e6)
angle = floats(min_value=-pi, max_value=pi)
dimension = integers(min_value=4, max_value=1024)
small_dimension = integers(min_value=4, max_value=128)
image_size = tuples(small_dimension, small_dimension)

# Stylo data
domain_values = arrays(np.float64, image_size, elements=real)
shape_mask = arrays(np.bool_, image_size)



from stylo.domain import RealDomain
from stylo.domain._factory import define_domain_transform
from stylo.shape import Shape

RealDomainTransform = define_domain_transform(RealDomain)


def find_base_transform(transform):
    """Given a domain transform, find its base class"""

    for base in transform.__bases__:

        if hasattr(base, "domain"):
            return base

    name = transform.__name__
    raise TypeError("{} is not a domain transform".format(name))


def find_base_domain(base_transform):
    """Given a base domain transform class, find the base domain that
    it transform."""

    for base in base_transform.__bases__:

        if hasattr(base, "_parameters") and not hasattr(base, "domain"):
            return base

    name = base_transform.__name__
    raise TypeError("{} is not a base domain transform".format(name))


class DomainTransformer:
    """A class used to apply transformations, it enables us to do the syntactic
    nicities when applying domain transforms."""

    def __init__(self, transform, *args, **kwargs):
        self.transform = transform
        self.args = args
        self.kwargs = kwargs

        self.base_transform = find_base_transform(transform)
        self.base_domain = find_base_domain(self.base_transform)
        self.name = transform.__name__.lower

    def __rrshift__(self, other):
        return self.apply_transform(other)

    def apply_transform(self, obj):

        transform = self.transform
        args = self.args
        kwargs = self.kwargs

        if isinstance(obj, (self.base_domain,)):
            return transform(obj, *args, **kwargs)

        if isinstance(obj, (Shape,)):
            obj._add_transform(lambda domain: self.transform(domain, *args, **kwargs))
            return obj

        obj_name = obj.__class__.__name__
        message = "Unable to perform a {} to a {}."
        raise TypeError(message.format(self.name, obj_name))


def define_transform(transform):
    def transform_func(*args, **kwargs):
        return DomainTransformer(transform, *args, **kwargs)

    return transform_func



import base64
import io
import PIL as P
import matplotlib.pyplot as plt

from abc import ABC, abstractmethod


class Drawable:
    def __init__(self, domain, shape, color):
        self.domain = domain
        self.shape = shape
        self.color = color

    def __iter__(self):
        return iter([self.domain, self.shape, self.color])


class Image(ABC):
    def __call__(
        self, width, height, filename=None, plot_size=None, encode=None, preview=True
    ):

        self.data = self._render(width, height)

        if filename:
            self.save(filename)
            return

        if encode:
            return self.encode()

        if preview:
            return self.preview(plot_size)

    def preview(self, plot_size):
        """Generate a matplotlib plot of the image data which can then be viewed from
        a Jupyter Notebook.

        :param plot_size: This controls how large the generated plot is
        :type plot_size: int

        :returns: A matplotlib AxesImage object.
        """

        if plot_size is None:
            plot_size = 12

        fig, ax = plt.subplots(1, figsize=(plot_size, plot_size))

        # Hide the axis - show just the image
        fig.axes[0].get_yaxis().set_visible(False)
        fig.axes[0].get_xaxis().set_visible(False)

        # Draw the image
        ax.imshow(self.data)

        # Return just the axis, jupyter notebooks will capture the figure and
        # display it anyway.
        return ax

    def encode(self):
        """Encode the image as a PNG represented by a base64 string.

        :return: The image encoded as a base64 string.
        """

        image = self._to_pil_image()

        with io.BytesIO() as byte_stream:
            image.save(byte_stream, "PNG")
            image_bytes = byte_stream.getvalue()

        return base64.b64encode(image_bytes)

    def save(self, filename):
        """Save the image to file as a PNG image.

        If the given filename already exists the existing image will be overwritten.

        :param filename: The file to save the image to.
        :type filename: str
        """

        image = self._to_pil_image()

        with open(filename, "wb") as f:
            image.save(f)

    def _to_pil_image(self):
        """Convert the numpy representation of the image into a PIL.Image object."""

        height, width, _ = self.data.shape
        return P.Image.frombuffer("RGB", (width, height), self.data, "raw", "RGB", 0, 1)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @abstractmethod
    def _render(self, width, height):
        pass


def render_drawable(drawable, image_data):
    """Given a drawable, render it onto the given image data."""

    domain, shape, color = drawable
    domain = shape._apply_transform(domain)

    height, width, depth = image_data.shape

    parameters = shape.parameters
    values = domain[parameters](width, height)

    coords = {k: v for k, v in zip(parameters, values)}
    mask = shape(**coords)

    return color(mask, image_data=image_data)



from .rectangular import RectangularDomain


class SquareDomain(RectangularDomain):
    """ A domain of the form :math:`[a, b] \\times [a, b] \\subset \\mathbb{R}^2`."""

    def __init__(self, x_min, x_max):
        super().__init__(x_min, x_max, x_min, x_max)

    @RectangularDomain.xmin.setter
    def xmin(self, value):
        RectangularDomain.xmin.fset(self, value)
        RectangularDomain.ymin.fset(self, value)

    @RectangularDomain.ymin.setter
    def ymin(self, value):
        RectangularDomain.ymin.fset(self, value)
        RectangularDomain.xmin.fset(self, value)

    @RectangularDomain.xmax.setter
    def xmax(self, value):
        RectangularDomain.xmax.fset(self, value)
        RectangularDomain.ymax.fset(self, value)

    @RectangularDomain.ymax.setter
    def ymax(self, value):
        RectangularDomain.ymax.fset(self, value)
        RectangularDomain.xmax.fset(self, value)


class UnitSquare(SquareDomain):
    """The domain :math:`[0, 1] \\times [0, 1] \\subset \\mathbb{R}^2`."""

    def __init__(self):
        super().__init__(0, 1)

    @SquareDomain.xmax.setter
    def xmax(self, value):
        raise AttributeError("can't set attribute")

    @SquareDomain.xmin.setter
    def xmin(self, value):
        raise AttributeError("can't set attribute")

    @SquareDomain.ymin.setter
    def ymin(self, value):
        raise AttributeError("can't set attribute")

    @SquareDomain.ymax.setter
    def ymax(self, value):
        raise AttributeError("can't set attribute")



from .clock import Clock
from stylo.utils import get_parameters


class Event:
    """A representation of an event that lasts for some duration."""

    def __init__(self, duration, name=None, fps=25, timeline=None):

        self._clock = Clock(duration, fps)
        self._timeline = timeline

        self.duration = duration
        self.name = "Event" if name is None else name

    def __repr__(self):
        return "{0.name}: {0.duration}s".format(self)

    def __call__(self, f):
        """Use this to register a time dependent function with this event."""
        return TimeDependent(self._clock, f)

    @property
    def started(self):
        return self._clock.started

    @property
    def finished(self):
        return self._clock.finished

    def start(self):
        self._clock.start()
        return self

    def tick(self):
        self._clock.tick()
        return self


class TimeDependent:
    """This wraps a callable so that it is given the appropriate time value."""

    def __init__(self, clock, f):

        self._params = get_parameters(f)

        if not set(self._params) <= set("ft"):
            raise TypeError("Expected time dependent function.")

        self.f = f
        self.clock = clock

    def __repr__(self):
        return "{0}\n{1}".format(repr(self.f), repr(self.clock))

    def __call__(self):
        values = self.clock[self._params]
        return self.f(*values)



from abc import ABC, abstractmethod

from stylo.color.colorspaces import ColorSpace, RGB8


class ColorMap(ABC):
    def __init__(self, colorspace=None):

        if colorspace is None:
            colorspace = RGB8

        self.colorspace = colorspace

    def __call__(self, shape, image_data=None):
        return self._paint(shape, image_data)

    def _parse_color(self, color):
        return self.colorspace.parse(color)

    @abstractmethod
    def _paint(self, shape, image_data):
        pass

    @property
    def colorspace(self):
        return self._colorspace

    @colorspace.setter
    def colorspace(self, value):

        if not issubclass(value, (ColorSpace,)):
            raise TypeError("Expected ColorSpace")

        self._colorspace = value



import numpy as np


class PolarConversion:
    """A helper class for defining new domain classes.

    Say you are implementing a new class, perhaps the :code:`Translation` domain
    transform where the interesting part of that class is implementing the shift in
    :code:`x` or the shift in :code:`y`. The implementation of the :code:`r` and
    :code:`t` is then done with respect to the shifted :code:`x` and :code:`y`

    This class provides implementations of the :code:`_get_rs` and :code:`_get_ts` that
    handle this conversion for you automatically, leaving you to focus on implementing
    the transform that interest you.

    To use this class, simply include it in your :code:`class` definition as follows.

    .. code-block:: python

       class MyDomain(PolarConversion, RealDomain):

           def _get_xs(self):
              ...

            def _get_ys(self):
               ...

    .. note::

       The order in which the classes are listed is *very* important.
    """

    def _get_r(self):
        """The conversion to the radial component in terms of :math:`x` and :math:`y`.

        .. math::

           r = \\sqrt{x^2 + y^2}
        """

        xs = self._get_x()
        ys = self._get_y()

        def mk_rs(width, height):

            x = xs(width, height)
            y = ys(width, height)

            return np.sqrt(x * x + y * y)

        return mk_rs

    def _get_t(self):
        """The conversion to the angular component in terms of :math:`x` and :math:`y`.

        .. math::

           \\theta = \\tan^{-1}{\\left(\\frac{y}{x}\\right)}
        """

        xs = self._get_x()
        ys = self._get_y()

        def mk_ts(width, height):

            x = xs(width, height)
            y = ys(width, height)

            return np.arctan2(y, x)

        return mk_ts


class CartesianConversion:
    """A helper class for defining new domain classes.

    Say you are implementing a new class, perhaps the :code:`Rotation` domain transform
    where the interesting part is the adjustments to the :code:`r` and :code:`t`
    coordinates. The implementation of :code:`x` and :code:`y` is then done with respect
    to the transformed :code:`r` and :code:`t` variables.

    This class provides implementations of the :code:`_get_xs` and :code:`_get_ys`
    methods to handle this conversion for you automatically, leaving you to focus on
    implementing the transform that interest you.

    To use this class, simply include it in your :code:`class` definition as follows

    .. code-block:: python

       class MyDomain(CartesianConversion, RealDomain):

           def _get_xs(self):
               ...

           def _get_ys(self):
               ...

    .. note::

       The order in which the classes are listed is *very* important.
    """

    def _get_x(self):
        """The conversion to the :math:`x` component in terms of :math:`r` and
        :math:`\\theta`

        .. math::

           x = r\\cos{(\\theta)}

        """

        rs = self._get_r()
        ts = self._get_t()

        def mk_xs(width, height):

            r = rs(width, height)
            t = ts(width, height)

            return r * np.cos(t)

        return mk_xs

    def _get_y(self):
        """The conversion to the :math:`y` component in terms of :math:`r` and
        :math:`\\theta`

        .. math::

           y = r\\sin{(\\theta)}

        """

        rs = self._get_r()
        ts = self._get_t()

        def mk_ys(width, height):

            r = rs(width, height)
            t = ts(width, height)

            return r * np.sin(t)

        return mk_ys



import numpy as np
from stylo.shape.shape import Shape


class ImplicitXY(Shape):
    """Define a curve using an implicit function in the :math:`x` and :math:`y`
    coordinates.

    A point :math:`(x, y)` lies on the curve defined by the function
    :math:`f(x, y)` if the following holds.

    .. math::

       f(x, y) = 0

    However since paths are one dimensional objects and have no area, if we used
    this rule as it stands we would never see the curve in question. So instead we
    introduce some error margin :math:`e > 0` and say that a point is on the curve
    if the following holds.

    .. math::

       |f(x, y)| \\leq e
    """

    def __init__(self, f, pt=0.01, above=False, below=False):
        """Construct an instance of :code:`ImplicitXY` according to the given
        parameters.

        :param f: The function that defines the curve.
        :param pt: The error margin, corresponds to :math:`e` in the above.
                   (Default: 0.01)
        :param above: If :code:`True` draw the area above the curve as a shaded
                      region (Default: :code:`False`)
        :param below: If :code:`True` draw the area underneath the curve as a shaded
                      region (Default: False)
        """
        self.f = f
        self.pt = pt
        self.above = above
        self.below = below

    def draw(self):

        if self.below:

            def below_curve(x, y):
                return self.f(x, y) >= 0

            return below_curve

        if self.above:

            def above_curve(x, y):
                return self.f(x, y) <= 0

            return above_curve

        def curve(x, y):
            error = self.f(x, y)
            return np.abs(error) <= self.pt

        return curve



import pytest
import numpy as np
from unittest import TestCase
from hypothesis import given, assume


from stylo.domain import SquareDomain
from stylo.shape import Ellipse, Circle, Rectangle, Square, Triangle, Line
from stylo.testing.shape import BaseShapeTest
from stylo.testing.strategies import real


@pytest.mark.shape
class TestEllipse(TestCase, BaseShapeTest):
    """Tests for the :code:`Ellipse` shape."""

    def setUp(self):
        self.shape = Ellipse()

    def test_init_defaults(self):
        """Ensure that :code:`Ellipse` is initialised with sane arguments by
        default."""

        ellipse = Ellipse()

        assert ellipse.x == 0
        assert ellipse.y == 0
        assert ellipse.a == 2
        assert ellipse.b == 1
        assert ellipse.r == 0.5
        assert ellipse.pt == 0.01

        assert not ellipse.fill

    def test_draw_pt(self):
        """Ensure that we can draw a complete ellipse as a curve."""

        ellipse = Ellipse(pt=0.1)
        domain = SquareDomain(-1, 1)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False, False],
                [False, False, False, False, False, False, False, False],
                [False,  True,  True,  True,  True,  True,  True, False],
                [ True, False, False, False, False, False, False,  True],
                [ True, False, False, False, False, False, False,  True],
                [False,  True,  True,  True,  True,  True,  True, False],
                [False, False, False, False, False, False, False, False],
                [False, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on

        assert (ellipse(domain, 8, 8) == expected).all()

    def test_draw_fill(self):
        """Ensure that we a draw a complete ellipse as a shaded region."""

        ellipse = Ellipse(fill=True)
        domain = SquareDomain(-1, 1)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False, False],
                [False, False, False, False, False, False, False, False],
                [False, False,  True,  True,  True,  True, False, False],
                [False,  True,  True,  True,  True,  True,  True, False],
                [False,  True,  True,  True,  True,  True,  True, False],
                [False, False,  True,  True,  True,  True, False, False],
                [False, False, False, False, False, False, False, False],
                [False, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on

        assert (ellipse(domain, 8, 8) == expected).all()


@pytest.mark.shape
class TestCircle(TestCase, BaseShapeTest):
    """Tests for the :code:`Circle` shape."""

    def setUp(self):
        self.shape = Circle()

    def test_init_defaults(self):
        """Ensure that :code:`Circle` is initialised with sane defaults."""

        circle = Circle()

        assert circle.x == 0
        assert circle.y == 0
        assert circle.r == 0.5

        assert circle.pt == 0.01
        assert not circle.fill

    def test_draw_pt(self):
        """Ensure that we can draw a :code:`Circle` as a curve."""

        circle = Circle(r=0.8, pt=0.15)
        domain = SquareDomain(-1, 1)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False, False],
                [False, False, True,  True,  True,   True, False, False],
                [False,  True, False, False, False, False,  True, False],
                [False,  True, False, False, False, False,  True, False],
                [False,  True, False, False, False, False,  True, False],
                [False,  True, False, False, False, False,  True, False],
                [False, False,  True,  True,  True,  True, False, False],
                [False, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on

        assert (expected == circle(domain, 8, 8)).all()

    def test_draw_full(self):
        """Ensure that we can draw a :code:`Circle` as a shaded region."""

        circle = Circle(r=0.8, fill=True)
        domain = SquareDomain(-1, 1)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False, False],
                [False, False, False,  True,  True, False, False, False],
                [False, False,  True,  True,  True,  True, False, False],
                [False,  True,  True,  True,  True,  True,  True, False],
                [False,  True,  True,  True,  True,  True,  True, False],
                [False, False,  True,  True,  True,  True, False, False],
                [False, False, False,  True,  True, False, False, False],
                [False, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on

        assert (expected == circle(domain, 8, 8)).all()


@pytest.mark.shape
class TestRectangle(TestCase, BaseShapeTest):
    """Tests for the :code:`Rectangle` shape."""

    def setUp(self):
        self.shape = Rectangle(0, 0, 0.6, 0.3)


@pytest.mark.shape
class TestSquare(TestCase, BaseShapeTest):
    """Tests for the :code:`Square` shape."""

    def setUp(self):
        self.shape = Square(0, 0, 0.75)


@pytest.mark.shape
class TestTriangle(TestCase, BaseShapeTest):
    """Tests for the :code:`Triangle` shape."""

    def setUp(self):
        self.shape = Triangle((1, 0.5), (0.2, 1), (0.4, 0.5))


@pytest.mark.shape
class TestLine(TestCase, BaseShapeTest):
    """Tests for the :code:`Line` shape."""

    def setUp(self):
        self.shape = Line()

    def test_init_defaults(self):
        """Ensure that the :code:`Line` is created with sane defaults."""

        line = Line()

        assert line.p1 == (0, 0)
        assert line.p2 == (1, 1)
        assert line.pt == 0.01
        assert not line.extend

    @given(x1=real, y1=real, x2=real, y2=real)
    def test_init_sorts_points(self, x1, y1, x2, y2):
        """Ensure that the :code:`Line` object sorts the points by the x coordinate."""

        assume(x1 != x2)

        line = Line((x1, y1), (x2, y2))
        assert line.p1[0] < line.p2[0]

    def test_draw_segment_pt(self):
        """Ensure we can draw a line segment."""

        line = Line(pt=0.1)
        domain = SquareDomain(-1, 1)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False, False],
                [False, False, False, False, False, False,  True, False],
                [False, False, False, False, False,  True, False, False],
                [False, False, False, False,  True, False, False, False],
                [False, False, False, False, False, False, False, False],
                [False, False, False, False, False, False, False, False],
                [False, False, False, False, False, False, False, False],
                [False, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on

        assert (expected == line(domain, 8, 8)).all()

    def test_draw_line_pt(self):
        """Ensure that we can draw a line that's been extended to infinity."""

        line = Line(pt=0.1, extend=True)
        domain = SquareDomain(-1, 1)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False,  True],
                [False, False, False, False, False, False,  True, False],
                [False, False, False, False, False,  True, False, False],
                [False, False, False, False,  True, False, False, False],
                [False, False, False,  True, False, False, False, False],
                [False, False,  True, False, False, False, False, False],
                [False,  True, False, False, False, False, False, False],
                [ True, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on

        assert (expected == line(domain, 8, 8)).all()

    def test_draw_line_below(self):
        """Ensure that we can draw a shaded region underneath a line."""

        line = Line(pt=0.1, extend=True, below=True)
        domain = SquareDomain(-1, 1)

        # fmt: off
        expected = np.array(
            [
                [False, False, False, False, False, False, False,  True],
                [False, False, False, False, False, False, False,  True],
                [False, False, False, False, False, False,  True,  True],
                [False, False, False, False, False,  True,  True,  True],
                [False, False, False, False,  True,  True,  True,  True],
                [False, False, False,  True,  True,  True,  True,  True],
                [False, False,  True,  True,  True,  True,  True,  True],
                [ True,  True,  True,  True,  True,  True,  True,  True],
            ]
        )
        # fmt: on
        assert (expected == line(domain, 8, 8)).all()

    def test_draw_line_above(self):
        """Ensure that we can draw a shaded region above a line."""

        line = Line(pt=0.1, extend=True, above=True)
        domain = SquareDomain(-1, 1)

        # fmt: off
        expected = np.array(
            [
                [True,  True,  True,  True,  True,  True,  True,  True],
                [True,  True,  True,  True,  True,  True,  True, False],
                [True,  True,  True,  True,  True,  True, False, False],
                [True,  True,  True,  True,  True, False, False, False],
                [True,  True,  True,  True, False, False, False, False],
                [True,  True,  True, False, False, False, False, False],
                [True,  True, False, False, False, False, False, False],
                [True, False, False, False, False, False, False, False],
            ]
        )
        # fmt: on
        assert (expected == line(domain, 8, 8)).all()



import base64
from hypothesis import given

from stylo.testing.strategies import dimension


class BaseImageTest:
    """A base class for writing :code:`Image` implementations.

    When writing your test case for a new :code:`Image` implementation you need to
    declare it as follows.

    .. code-block:: python

       from unittest import TestCase
       from stylo.testing.image import BaseImageTest

       class TestMyImage(TestCase, BaseImageTest):
           ...

    .. note::

       The order in which you write the classes is *very* important.

    You also need to define the :code:`setUp` method to set the :code:`image` attribute
    equal to an instance of your image implementation.

    .. code-block:: python

       def setUp(self):
           self.image = MyImage()

    Then in addition to any tests you write, your :code:`Image` implementation will be
    automatically tested to see if passes the checks defined below.
    """

    @given(width=dimension, height=dimension)
    def test_encode(self, width, height):
        """Ensure that if the :code:`encode=True` keyword argument is given then a
        base64 encoded string representing the image in PNG format is returned."""

        # Every PNG image starts with the same magic number.
        # https://en.wikipedia.org/wiki/Portable_Network_Graphics
        magic = base64.b64encode(bytes.fromhex("89504e470d0a1a0a"))

        image_bytes = self.image(width, height, encode=True)

        assert isinstance(image_bytes, (bytes,)), "Expected byte string"
        assert magic[0:11] == image_bytes[0:11]



from math import floor
from pathlib import Path

from tqdm import tqdm
from .clock import Clock


class Timeline:
    """All time begins here."""

    def __init__(self, seconds=None, minutes=None, hours=None, fps=25):
        self._filename = None
        self._clock = Clock(fps)
        duration = 0

        if hours is not None:
            duration += hours * 60 * 60

        if minutes is not None:
            duration += minutes * 60

        if seconds is not None:
            duration += seconds

        self.fps = fps
        self.framecount = floor(duration * fps)
        self.duration = duration

    def _build_filename(self, filename):
        path = Path(filename)

        pad = "-{{:0{}d}}".format(len(str(self.framecount)))
        fname = path.stem + pad + path.suffix

        fmt_string = str(path.parent.joinpath(fname))
        self._filename = fmt_string

    @property
    def filename(self):
        if self._filename is not None:
            return self._filename.format(self._clock.tick)

    def main(self, f):
        self._clock.on_tick(f)
        return f

    def render(self, filename=None):

        if filename is not None:
            self._build_filename(filename)

        with tqdm(total=self.framecount) as pbar:

            while self._clock.tick < self.framecount:
                self._clock()
                pbar.update(1)



from stylo.design import Position, Trajectory
from ._param_factory import (
    define_parameter_group_test,
    define_time_dependent_parameter_group_test,
)

BasePositionTest = define_parameter_group_test(Position)
BaseTrajectoryTest = define_time_dependent_parameter_group_test(Trajectory)



#!/usr/bin/env python
"""
This script aims to provide similar functionality to the :code:`sphinx-apidoc` command
but produce results with better integration with the way the documentation is structured.

Usage
-----

As it currently stands this script depends only on the Python standard library. If you
are making changes to the script and want to run it in isolation then it can be
done as follows:

::

   $ python scripts/apidoc.py -m <path_to_module> -o <path_to_output>


Todo
----

Here are some changes that would be good to make at some point

- Generalise this script so that it can be applied to code in the tests/ folder
- Add an option to ignore certain files

"""
import os
from os import path

import argparse
import pathlib

RST_TEMPLATE = """

.. {reference}:

{title}

.. automodule:: {module}
   :members:
   :undoc-members:
   :private-members:
   :show-inheritance:

"""


RST_INDEX = """

.. _api_reference:

API Reference
=============

.. toctree::

   {}
"""


def format_title(title):
    """Given a module name, format it as a top level rst heading.

    This function will capitalize each word in the module name and return it with
    the :code:`=` underline.

    :param title: The module name
    :type title: str
    """

    underline = "=" * len(title)
    rst_title = ".".join(s.capitalize() for s in title.split("."))

    return "{0}\n{1}".format(rst_title, underline)


def format_rst(name):
    """Given the module name, format the rst contents that will be written to file.

    :param name: The module name
    :type name: str
    """

    context = {
        "title": format_title(name),
        "module": name,
        "reference": "_" + name.replace(".", "_"),
    }

    return RST_TEMPLATE.format(**context)


def find_modules(module_dir):
    """Given a folder, recursively find all the module files to right the documentation
    for.

    :param module_dir: The folder to start the search in
    :type module_dir: str
    """

    root = pathlib.Path(module_dir)
    modules = []

    for path in root.glob("**/*.py"):

        if path.name == "__init__.py":
            path = path.parent

        module = str(path).replace(".py", "").replace("/", ".")
        modules.append(module)

    return modules


def write_rst(root, name, rst, index):
    """Write the given rst to file."""

    fname = name.replace(".", "_").replace("__", "_")
    index.append(fname)

    filename = path.join(root, fname) + ".rst"

    dir = pathlib.Path(filename).parent

    if not dir.exists():
        os.mkdir(str(dir))

    with open(filename, "w") as f:
        f.write(rst)

    # print("\t{}.".format(name))


def write_index(out, index):

    entries = "\n   ".join(index)

    with open(path.join(out, "index.rst"), "w") as f:
        f.write(RST_INDEX.format(entries))


def process_arguments():

    parser = argparse.ArgumentParser(prog="apidoc.py")

    parser.add_argument(
        "-o, --output",
        nargs=1,
        type=str,
        dest="output",
        help="the folder to write the rst files to.",
    )

    parser.add_argument(
        "-m, --module",
        nargs=1,
        type=str,
        dest="module",
        required=True,
        help="the module to document.",
    )

    return parser.parse_args()


def main():

    args = process_arguments()
    out = args.output[0]

    modules = find_modules(args.module[0])
    index = []

    for module in modules:

        rst = format_rst(module)
        write_rst(out, module, rst, index)

    write_index(out, sorted(index))

    print("Documented {} modules".format(len(modules)))


if __name__ == "__main__":
    main()



import pytest
import numpy as np
from unittest import TestCase
from hypothesis import given

from stylo.domain import UnitSquare
from stylo.shape import Circle, Shape
from stylo.shape.shape import InvertedShape, ANDedShape, ORedShape, XORedShape
from stylo.testing.strategies import dimension
from stylo.testing.shape import BaseShapeTest


class MyShape(Shape):
    """A shape that depends on something other than x and y. This should be replaced
    with a shape from the library when such a shape is available.
    """

    def draw(self):
        return lambda r, t: r < t


@pytest.mark.shape
class TestInvertedShape(TestCase, BaseShapeTest):
    """Tests for the :code:`InvertedShape` class."""

    def setUp(self):
        self.baseshape = Circle(0, 0, 0.8)
        self.shape = InvertedShape(self.baseshape)

        self.domain = UnitSquare()

    def test_exposes_parameters(self):
        """Ensure that the :code:`InvertedShape` class exposes the parameters defined
        by the underlying shape.
        """
        assert self.shape.parameters == self.baseshape.parameters

    @given(width=dimension, height=dimension)
    def test_shape_is_inverted(self, width, height):
        """Ensure that an :code:`InvertedShape` is inverted by checking that a shape
        anded with its inversion should be False everywhere.
        """

        mask = self.baseshape(self.domain, width, height)
        inverted_mask = self.shape(self.domain, width, height)

        assert np.logical_not(np.logical_and(mask, inverted_mask)).all()

    @given(width=dimension, height=dimension)
    def test_double_inversion(self, width, height):
        """Ensure that an inverted :code:`InvertedShape` is the same as the shape
        itself.
        """

        double_inversion = InvertedShape(self.shape)

        mask = self.baseshape(self.domain, width, height)
        inverted_mask = double_inversion(self.domain, width, height)

        assert (mask == inverted_mask).all()

    def test_draw_method(self):
        """Rightly or wrongly, :code:`InvertedShape` currently doesn't make use of the
        draw method. So we'll override and disable this test from the base class."""
        pass


@pytest.mark.shape
class TestANDedShape(TestCase, BaseShapeTest):
    """Tests for the :code:`ANDedShape` class."""

    def setUp(self):
        self.a = Circle(0, 0, 0.8)
        self.not_a = InvertedShape(self.a)
        self.b = MyShape()

        self.shape = ANDedShape(self.a, self.not_a)
        self.domain = UnitSquare()

    def test_exposes_parameters(self):
        """Ensure that the parameters needed to define both a and b are exposed."""

        composite = ANDedShape(self.a, self.b)

        for p in self.a.parameters:
            assert p in composite.parameters, "Expected parameter: {}".format(p)

        for p in self.b.parameters:
            assert p in composite.parameters, "Expected parameter: {}".format(p)

    @given(width=dimension, height=dimension)
    def test_shapes_are_anded(self, width, height):
        """Ensure that the shapes are anded together. This is done by checking that the
        circle and its inversion results in False everywhere."""

        mask = self.shape(self.domain, width, height)

        assert np.logical_not(mask).all()

    def test_draw_method(self):
        """Rightly or wrongly, :code:`ANDedShape` currently doesn't make use of the
        draw method. So we'll override and disable this test from the base class."""
        pass


@pytest.mark.shape
class TestORedShape(TestCase, BaseShapeTest):
    """Tests for the :code:`ORedShape` class."""

    def setUp(self):
        self.a = Circle(0, 0, 0.8)
        self.not_a = InvertedShape(self.a)
        self.b = MyShape()

        self.shape = ORedShape(self.a, self.not_a)
        self.domain = UnitSquare()

    def test_exposes_parameters(self):
        """Ensure that the parameters needed to define both a and b are exposed."""

        composite = ORedShape(self.a, self.b)

        for p in self.a.parameters:
            assert p in composite.parameters, "Expected parameter: {}".format(p)

        for p in self.b.parameters:
            assert p in composite.parameters, "Expected parameter: {}".format(p)

    @given(width=dimension, height=dimension)
    def test_shapes_are_ored(self, width, height):
        """Ensure that the shapes are ored together. This is done by checking that the
        circle and its inversion results in True everywhere."""

        mask = self.shape(self.domain, width, height)

        assert mask.all()

    def test_draw_method(self):
        """Rightly or wrongly, :code:`ORedShape` currently doesn't make use of the
        draw method. So we'll override and disable this test from the base class."""
        pass


@pytest.mark.shape
class TestXORedShape(TestCase, BaseShapeTest):
    """Tests for the :code:`XORedShape` class."""

    def setUp(self):
        self.a = Circle(0, 0, 0.8)
        self.not_a = InvertedShape(self.a)
        self.b = MyShape()

        self.shape = XORedShape(self.a, self.not_a)
        self.domain = UnitSquare()

    def test_exposes_parameters(self):
        """Ensure that the parameters needed to define both a and b are exposed."""

        composite = XORedShape(self.a, self.b)

        for p in self.a.parameters:
            assert p in composite.parameters, "Expected parameter: {}".format(p)

        for p in self.b.parameters:
            assert p in composite.parameters, "Expected parameter: {}".format(p)

    @given(width=dimension, height=dimension)
    def test_shapes_are_xored(self, width, height):
        """Ensure that the shapes are xored together. This is done by checking that the
        circle and its inversion results in True everywhere."""

        mask = self.shape(self.domain, width, height)

        assert mask.all()

    @given(width=dimension, height=dimension)
    def test_shapes_are_xored_part2(self, width, height):
        """Ensure that the shapes are xored together. This is done by checking that the
        circle and itself results in False everywhere."""

        composite = XORedShape(self.a, self.a)
        mask = composite(self.domain, width, height)

        assert np.logical_not(mask).all()

    def test_draw_method(self):
        """Rightly xor wrongly, :code:`XORedShape` currently doesn't make use of the
        draw method. So we'll override and disable this test from the base class."""
        pass



from unittest import TestCase

import pytest
import numpy as np
import numpy.testing as npt
from hypothesis import given, assume, example

from stylo.domain.rectangular import RectangularDomain, get_real_domain
from stylo.testing.strategies import real, dimension
from stylo.testing.domain import BaseRealDomainTest


@pytest.mark.domain
class TestRectangularDomain(TestCase, BaseRealDomainTest):
    """Tests for the RectangularDomain class"""

    def setUp(self):
        self.domain = RectangularDomain(-1, 1, -1, 1)

    @given(xmin=real, xmax=real)
    def test_init_checks_x_min_max(self, xmin, xmax):
        """Ensure that the __init__ method checks to see that the value
        of :code:`xmin < xmax`."""

        assume(xmax <= xmin)

        with pytest.raises(ValueError) as err:
            RectangularDomain(xmin, xmax, 0, 1)

        self.assertIn("must be strictly less than", str(err.value))
        self.assertIn("xmin", str(err.value))
        self.assertIn("xmax", str(err.value))

    @given(ymin=real, ymax=real)
    def test_init_checks_y_min_max(self, ymin, ymax):
        """Ensure that the __init__ method checks to see that the value
        of :code:`ymin < ymax`."""

        assume(ymax <= ymin)

        with pytest.raises(ValueError) as err:
            RectangularDomain(0, 1, ymin, ymax)

        self.assertIn("must be strictly less than", str(err.value))
        self.assertIn("ymin", str(err.value))
        self.assertIn("ymax", str(err.value))

    @given(xmin=real)
    def test_xmin_property(self, xmin):
        """Ensure that the :code:`xmin` property works as expected.

        The :code:`xmin` property should:

        - Return the value of :code:`xmin`
        - Be used to set a new value of :code:`xmin`
        """

        domain = RectangularDomain(xmin, xmin + 1, 0, 1)
        self.assertEqual(xmin, domain.xmin)

        domain.xmin = xmin - 2
        self.assertEqual(xmin - 2, domain.xmin)

    def test_xmin_property_checks_value(self):
        """Ensure that the :code:`xmin` property checks the value it is
        given to make sure it makes sense.

        The new value should:

        - Be a numeric value
        - Be strictly less than the current value of :code:`xmax`
        """

        domain = RectangularDomain(0, 1, 0, 1)

        with pytest.raises(TypeError) as err:
            domain.xmin = "String"

        self.assertIn("must be a number", str(err.value))

        with pytest.raises(ValueError) as err:
            domain.xmin = 10

        self.assertIn("must be strictly less than", str(err.value))

    @given(xmax=real)
    def test_xmax_property(self, xmax):
        """Ensure that the :code:`xmax` property works as expected.

        The :code:`xmax` property should:

        - Return the value of :code:`xmax`
        - Be used to set a new value of :code:`xmax`
        """

        domain = RectangularDomain(xmax - 1, xmax, 0, 1)
        self.assertEqual(xmax, domain.xmax)

        domain.xmax = xmax + 2
        self.assertEqual(xmax + 2, domain.xmax)

    def test_xmax_property_checks_value(self):
        """Ensure that the :code:`xmax` property checks the value it is
        given to make sure it makes sense.

        The new value should:

        - Be a numeric value
        - Be strictly greater than the current value of :code:`xmin`
        """

        domain = RectangularDomain(0, 1, 0, 1)

        with pytest.raises(TypeError) as err:
            domain.xmax = "String"

        self.assertIn("must be a number", str(err.value))

        with pytest.raises(ValueError) as err:
            domain.xmax = -1

        self.assertIn("must be strictly larger than", str(err.value))

    @given(ymin=real)
    def test_ymin_property(self, ymin):
        """Ensure that the :code:`ymin` property as expected.

        The :code:`ymin` property should:

        - Return the value of :code:`ymin`
        - Be used to set a new value of :code:`ymin`
        """

        domain = RectangularDomain(0, 1, ymin, ymin + 1)
        self.assertEqual(ymin, domain.ymin)

        domain.ymin = ymin - 2
        self.assertEqual(ymin - 2, domain.ymin)

    def test_ymin_property_checks_value(self):
        """Ensure that the :code:`ymin` property checks the value
        it is given to make sure it makes sense.

        The new value should:

        - Be a numeric value
        - Be strictly less than the value of :code:`ymax`
        """

        domain = RectangularDomain(0, 1, 0, 1)

        with pytest.raises(TypeError) as err:
            domain.ymin = "String"

        self.assertIn("must be a number", str(err.value))

        with pytest.raises(ValueError) as err:
            domain.ymin = 10

        self.assertIn("must be strictly less than", str(err.value))

    @given(ymax=real)
    def test_ymax_property(self, ymax):
        """Ensure that the :code:`ymax` property works as expected.

        The :code:`ymax` property should

        - Return the current value of :code:`ymax`
        - Be used to set a new value of :code:`ymax`
        """

        domain = RectangularDomain(0, 1, ymax - 1, ymax)
        self.assertEqual(ymax, domain.ymax)

        domain.ymax = ymax + 1
        self.assertEqual(ymax + 1, domain.ymax)

    def test_ymax_property_checks_value(self):
        """Ensure that the :code:`ymax` property checks the value it is given
        to make sure it makes sense

        The new value should:

        - Be a numeric value
        - Be strictly greater than :code:`ymin`
        """

        domain = RectangularDomain(0, 1, 0, 1)

        with pytest.raises(TypeError) as err:
            domain.ymax = "String"

        self.assertIn("must be a number", str(err.value))

        with pytest.raises(ValueError) as err:
            domain.ymax = -1

        self.assertIn("must be strictly larger than", str(err.value))

    @given(xmin=real, xmax=real, width=dimension, height=dimension)
    def test_x_property_values(self, width, height, xmin, xmax):
        """Ensure that the :code:`x` property produces the expected values.

        The produced values should

        - Start with xmin
        - End with xmax
        - Be larger than xmin inbetween
        - Be less than xmax inbetween

        .. note::

           We don't have to check the shape of the produced array as that is
           handled by the base class.
        """
        assume(xmin < xmax)
        domain = RectangularDomain(xmin, xmax, 0, 1)

        XS = domain.x(width, height)

        npt.assert_array_equal(XS[:0], xmin)
        npt.assert_array_equal(XS[:, -1], xmax)

        interior = XS[:, 1:-1]
        self.assertTrue(
            (interior <= xmax).all(), "The interior should be less than xmax"
        )
        self.assertTrue(
            (xmin <= interior).all(), "The interior should be greater then xmin"
        )

    @given(ymin=real, ymax=real, width=dimension, height=dimension)
    def test_y_property_values(self, width, height, ymin, ymax):
        """Ensure that the :code:`y` property produces the expected values.

        The produced values should

        - Start with ymax
        - End with ymin
        - Be larger than ymin inbetween
        - Be less than ymax inbetween

        .. note::

           We don't have to check the shape of the produced array as that is
           handled by the base class
        """
        assume(ymin < ymax)
        domain = RectangularDomain(0, 1, ymin, ymax)

        YS = domain.y(width, height)

        npt.assert_array_equal(YS[0, :], ymax)
        npt.assert_array_equal(YS[-1, :], ymin)

        interior = YS[1:-1,]

        self.assertTrue(
            (interior <= ymax).all(), "The interior should be less than ymax"
        )
        self.assertTrue(
            (ymin <= interior).all(), "The interior should be greater than ymin"
        )

    @given(width=dimension, height=dimension)
    def test_r_property_values(self, width, height):
        """Ensure that the :code:`r` property produces the expected values.

        The produced values should

        - All be greater than zero
        - Be radially symmetric but how do we test that??
        """

        domain = RectangularDomain(0, 1, 0, 1)

        RS = domain.r(width, height)
        self.assertTrue((RS >= 0).all(), "Values of r should be strictly positive")

    @given(width=dimension, height=dimension)
    def test_t_property_values(self, width, height):
        """Ensure that :code:`t` property produces the expected values.

        The produced values should

        - Be between -pi and pi
        - Be the same along "radial lines" but how do we test that??
        """

        domain = RectangularDomain(0, 1, 0, 1)

        TS = domain.t(width, height)
        self.assertTrue(
            (TS <= np.pi / 2).all(), "Angles should be less than or equal to pi/2"
        )
        self.assertTrue(
            (TS >= -np.pi / 2).all(), "Angles should be greater than or equal to -pi/2"
        )


@pytest.mark.domain
class TestGetRealDomain:
    """Tests for the :code:`get_real_domain` function."""

    @given(width=dimension)
    @example(width=0)
    def test_checks_width_parameter(self, width):
        """Ensure that an error is thrown if a zero or negative value for :code:`width`
        is given.
        """

        with pytest.raises(ValueError) as err:
            get_real_domain(-width, 4)

        assert "must be a positive number" in str(err.value)

    @given(height=dimension)
    @example(height=0)
    def test_checks_height_parameter(self, height):
        """Ensure that an error is thrown if a zero or negative value for :code:`height`
        is given.
        """

        with pytest.raises(ValueError) as err:
            get_real_domain(4, -height)

        assert "must be a positive number" in str(err.value)

    @given(scale=real)
    @example(scale=0)
    def test_checks_scale_positive(self, scale):
        """Ensure that an error is thrown if a zero or negative value for :code:`scale`
        is given.
        """
        assume(scale >= 0)

        with pytest.raises(ValueError) as err:
            get_real_domain(4, 4, scale=-scale)

        assert "must be strictly positive" in str(err.value)

    @given(width=dimension, height=dimension, scale=real)
    def test_ylength_is_scale(self, width, height, scale):
        """Ensure that the length of the y interval corresponds to the value of the
        scale parameter.
        """
        assume(scale > 0)
        domain = get_real_domain(width, height, scale)

        ymax = domain.ymax
        ymin = domain.ymin

        assert (ymax - ymin) == scale

    @given(width=dimension, height=dimension, scale=real)
    def test_domain_matches_aspect_ratio(self, width, height, scale):
        """Ensure that the aspect ratio of the domain matches the aspect ratio specified
        by the width and height
        """

        assume(scale > 0)
        domain = get_real_domain(width, height, scale)

        ylength = domain.ymax - domain.ymin
        xlength = domain.xmax - domain.xmin

        domain_ratio = xlength / ylength
        aspect_ratio = width / height

        assert domain_ratio == pytest.approx(aspect_ratio, 0.1)



import pytest
from hypothesis import given, assume
from unittest import TestCase

from stylo.domain.square import SquareDomain, UnitSquare
from stylo.testing.domain import BaseRealDomainTest
from stylo.testing.strategies import real


@pytest.mark.domain
class TestSquareDomain(TestCase, BaseRealDomainTest):
    """Tests for the :code:`SquareDomain` class"""

    def setUp(self):
        self.domain = SquareDomain(-1, 1)

    @given(value=real)
    def test_xmin_property_sets_ymin(self, value):
        """Ensure that the :code:`xmin` property also sets the :code:`ymin`
        property to maintain the square aspect ratio."""
        assume(value < 1)

        domain = SquareDomain(-1, 1)
        domain.xmin = value

        self.assertEqual(value, domain.xmin)
        self.assertEqual(value, domain.ymin)

    @given(value=real)
    def test_ymin_property_sets_xmin(self, value):
        """Ensure that the :code:`ymin` property also sets the :code:`xmin`
        property to maintain the square aspect ratio."""

        assume(value < 1)

        domain = SquareDomain(-1, 1)
        domain.ymin = value

        self.assertEqual(value, domain.xmin)
        self.assertEqual(value, domain.ymin)

    @given(value=real)
    def test_xmax_property_sets_ymax(self, value):
        """Ensure that the :code:`xmax` property also sets the :code:`ymax`
        property to maintain the square aspect ratio."""

        assume(value > 1)

        domain = SquareDomain(-1, 1)
        domain.xmax = value

        self.assertEqual(value, domain.xmax)
        self.assertEqual(value, domain.ymax)

    @given(value=real)
    def test_ymax_property_sets_xmax(self, value):
        """Ensure that the :code:`ymax` property also sets the :code:`xmax`
        property to maintain the square aspect ratio."""

        assume(value > 1)

        domain = SquareDomain(-1, 1)
        domain.ymax = value

        self.assertEqual(value, domain.xmax)
        self.assertEqual(value, domain.ymax)


@pytest.mark.domain
class TestUnitSquareDomain(TestCase, BaseRealDomainTest):
    """Tests for the :code:`UnitSquare` class"""

    def setUp(self):
        self.domain = UnitSquare()

    def test_init(self):
        """Ensure that the domain is initialised with the proper values to
        define the unit square"""

        domain = UnitSquare()

        self.assertEqual(0, domain.xmin)
        self.assertEqual(0, domain.ymin)

        self.assertEqual(1, domain.xmax)
        self.assertEqual(1, domain.ymax)

    def test_xmax_property_not_settable(self):
        """Ensure that the :code:`xmax` property cannot be set on the UnitSquare
        domain."""

        domain = UnitSquare()

        with pytest.raises(AttributeError) as err:
            domain.xmax = 2

        self.assertIn("can't set attribute", str(err.value))

    def test_xmin_property_not_settable(self):
        """Ensure that the :code:`xmin` property cannot be set on the UnitSquare
        domain."""

        domain = UnitSquare()

        with pytest.raises(AttributeError) as err:
            domain.xmin = -1

        self.assertIn("can't set attribute", str(err.value))

    def test_ymin_property_not_settable(self):
        """Ensure that the :code:`ymin` property cannot be set on the UnitSquare
        domain."""

        domain = UnitSquare()

        with pytest.raises(AttributeError) as err:
            domain.ymin = -1

        self.assertIn("can't set attribute", str(err.value))

    def test_ymax_property_not_settable(self):
        """Ensure that the :code:`ymax` property cannot be set on the UnitSquare
        domain."""

        domain = UnitSquare()

        with pytest.raises(AttributeError) as err:
            domain.ymax = 2

        self.assertIn("can't set attribute", str(err.value))



import numpy.testing as npt
import pytest
from hypothesis import given
from unittest import TestCase

from stylo.domain.square import UnitSquare
from stylo.domain.transform.translation import Translation
from stylo.testing.domain import BaseRealDomainTest
from stylo.testing.strategies import dimension, real


@pytest.mark.domain
class TestTranslation(TestCase, BaseRealDomainTest):
    """Tests for the Translation class"""

    def setUp(self):

        # We just need any instance of the translation to test the interface
        # with, we'll get to the details ourselves in the test cases here.
        self.domain = Translation(UnitSquare(), 1, 1)

    @given(width=dimension, height=dimension, dx=real)
    def test_dx_affects_x_values(self, width, height, dx):
        """Ensure that the :code:`dx` shifts each value by the negative of the value
        given by the user so that the translation is intuitive. Also ensure that only
        the :code:`x` values are affected."""

        domain = UnitSquare()
        xs, ys = domain["xy"](width, height)

        translated = Translation(domain, dx=dx, dy=0)
        t_xs, t_ys = translated["xy"](width, height)

        # The y values should be unaffected.
        npt.assert_array_equal(ys, t_ys)

        difference = t_xs - xs
        npt.assert_array_almost_equal(difference, -dx)

    @given(width=dimension, height=dimension, dy=real)
    def test_dy_affects_y_values(self, width, height, dy):
        """Ensure that the :code:`dy` shifts each value by the negative of the value
        given by the user so that the translation is intuitive. Also ensure that only
        the :code:`y` values are affected."""

        domain = UnitSquare()
        xs, ys = domain["xy"](width, height)

        translated = Translation(domain, dx=0, dy=dy)
        t_xs, t_ys = translated["xy"](width, height)

        # The x values should be unaffected
        npt.assert_array_equal(xs, t_xs)

        difference = t_ys - ys
        npt.assert_array_almost_equal(difference, -dy)



import numpy as np
from hypothesis import given

from stylo.domain import UnitSquare, RealDomain
from stylo.shape.shape import InvertedShape, ANDedShape, ORedShape, XORedShape
from stylo.testing.strategies import dimension
from stylo.utils import get_parameters


class BaseShapeTest:
    """A base class for testing :code:`Shape` implementations.

    When writing your test case for a new :code:`Shape` implementation you need to
    declare it as follows.

    .. code-block:: python

       from unittest import TestCase
       from stylo.testing.shape import BaseShapeTest

       class TestMyShape(TestCase, BaseShapeTest):
           ...

    .. note::

       The order in which you write the classes is *very* important.

    You also need to define the :code:`setUp` method to set the :code:`shape` attribute
    equal to an instance of your shape implementation.

    .. code-block:: python

       def setUp(self):
           self.shape = MyShape()

    Then in addition to any tests you write, your :code:`Shape` implementation will be
    automatically tested to see if it passes the checks defined below.
    """

    @given(width=dimension, height=dimension)
    def test_render_with_domain(self, width, height):
        """Ensure that a shape can be called with a domain and integers representing the
        width and height in pixels of an image. The result must:

        - Be numpy array
        - With shape :code:`(height, width)`
        - With elements of type bool.
        """
        domain = UnitSquare()
        mask = self.shape(domain, width, height)

        self.assertTrue(isinstance(mask, (np.ndarray,)), "Expected numpy array.")
        self.assertEqual((height, width), mask.shape)
        self.assertTrue(mask.dtype == np.bool, "Expected boolean array")

    @given(width=dimension, height=dimension)
    def test_render_with_args(self, width, height):
        """Ensure that a shape can be called with keyword arguments. Each of the
        arguments should be numpy arrays of the same shape with names corresponding to
        each of the coordinate names the shape is defined with. The result must:

        - Be a numpy array
        - With shape :code:`(height, width)`
        - With elements of type bool.
        """
        domain = UnitSquare()
        parameters = self.shape.parameters

        values = domain[parameters](width, height)
        params = {k: v for k, v in zip(parameters, values)}

        mask = self.shape(**params)

        self.assertTrue(isinstance(mask, (np.ndarray,)), "Expected numpy array.")
        self.assertEqual((height, width), mask.shape)
        self.assertTrue(mask.dtype == np.bool, "Expected boolean array.")

    def test_shape_can_be_inverted(self):
        """Ensure that a shape can be inverted by checking that an instance of
        :code:`InvertedShape` is returned when the :code:`~` operator is used on the
        shape.
        """

        inverted_shape = ~self.shape

        assert isinstance(inverted_shape, (InvertedShape,)), "Expected inverted shape."
        assert inverted_shape.shape == self.shape

    def test_shape_can_be_anded(self):
        """Ensure that a shape can be ANDed by checking that an instance of
        :code:`ANDedShape` is returned when the :code:`&` operator is used."""

        anded_shape = self.shape & self.shape

        assert isinstance(anded_shape, (ANDedShape,)), "Expected anded shape"
        assert anded_shape.a == self.shape
        assert anded_shape.b == self.shape

    def test_shape_can_be_ored(self):
        """Ensure that a shape can be ored by checking that an instance of
        :code:`ORedShape` is returned when the :code:`|` operator is used."""

        ored_shape = self.shape | self.shape

        assert isinstance(ored_shape, (ORedShape,)), "Expected ored shape"
        assert ored_shape.a == self.shape
        assert ored_shape.b == self.shape

    def test_shape_can_be_xored(self):
        """Ensure that a shape can be xored by checking that an instance of
        :code:`XORedShape` is returned when the :code:`^` operator is used."""

        xored_shape = self.shape ^ self.shape

        assert isinstance(xored_shape, (XORedShape,)), "Expected xored shape"
        assert xored_shape.a == self.shape
        assert xored_shape.b == self.shape

    def test_parameters_property(self):
        """Ensure that the :code:`parameters` property returns the coordinates that the
        shape is defined with. The parameters must

        - Be iterable
        - Be unique
        - Be declared on the :code:`RealDomain` interface.
        """

        coordinates = self.shape.parameters
        unique = set(coordinates)

        self.assertEqual(len(unique), len(coordinates))

        for c in coordinates:
            if c not in RealDomain._parameters:
                message = "{} is not a valid coordinate variable."
                raise ValueError(message.format(c))

    def test_draw_method(self):
        """Ensure that the :code:`draw` method returns a valid shape function. A shape
        function must

        - Be callable
        - Only take coordinate arguments defined in :code:`RealDomain`
        - These arguments must match those defined in the :code:`parameters` property.
        """

        method = self.shape.draw()

        self.assertTrue(callable(method), "Expected callable.")
        parameters = get_parameters(method)

        self.assertEqual(parameters, self.shape.parameters)

    def test_has_transforms(self):
        """Ensure that a shape has a :code:`_transforms` attribute that is equal to the
        empty list."""

        assert self.shape._transforms == []

    def test_add_transform(self):
        """Ensure that when transforms are added to the list that they are added to the
        front of the list."""

        shape = self.shape
        shape._add_transform(1)
        shape._add_transform(2)

        assert [2, 1] == shape._transforms

    def test_apply_transform_none(self):
        """Ensure that when a shape carries no transforms that the domain is returned
        untouched."""

        domain = UnitSquare()
        assert self.shape._apply_transform(domain) == domain

    def test_apply_transform(self):
        """Ensure that when a shape carries transforms that they are applied."""

        self.shape._add_transform(lambda n: 2 * n)
        self.shape._add_transform(lambda n: n + 1)

        assert 22 == self.shape._apply_transform(10)



from ._param_factory import (  # noqa: F401
    define_parameter_group,
    define_time_dependent_parameter_group,
)

Position = define_parameter_group("Position", "x, y")
Trajectory = define_time_dependent_parameter_group("Trajectory", "x, y")


class StaticPosition(Position):
    """A basic implementation of the :code:`Position` parameter group.

    It takes an x and y and returns them.
    """

    def __init__(self, x=0, y=0):
        self.x_pos = x
        self.y_pos = y
        super().__init__()

    def calculate(self):
        return {"x": self.x_pos, "y": self.y_pos}


def id_func(t):
    return t


class ParametricTrajectory(Trajectory):
    """A basic implementation of the :code:`Trajectory` time dependent parameter group.

    It takes two functions :code:`x(t)` and :code:`y(t)` and returns them.
    """

    def __init__(self, x=None, y=None):

        if x is None:
            x = id_func

        if y is None:
            y = id_func

        self.x_t = x
        self.y_t = y
        super().__init__()

    def calculate(self, t):
        return {"x": self.x_t(t), "y": self.y_t(t)}



from setuptools import setup, find_packages

from stylo import __version__


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="stylo",
    version=__version__,
    description="Using a blend of Python and Maths for the creation of images",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Multimedia :: Graphics",
    ],
    author="Alex Carney",
    author_email="alcarneyme@gmail.com",
    license="MIT",
    packages=find_packages(".", exclude="tests"),
    install_requires=["Pillow", "hypothesis", "matplotlib", "numpy", "pytest", "tqdm"],
    python_requires=">=3.5",
    setup_requires=["pytest-runner"],
    test_suite="tests",
    include_package_data=True,
    zip_safe=False,
)



from ._factory import define_domain

RealDomain = define_domain("RealDomain", "x,y,r,t")

from .rectangular import RectangularDomain, get_real_domain  # noqa: F401
from .square import SquareDomain, UnitSquare  # noqa: F401



import numpy as np

from stylo.domain import get_real_domain
from stylo.image.image import Drawable, Image, render_drawable


class SimpleImage(Image):
    def __init__(self, shape, color, background=None, scale=2):

        self.shape = shape
        self.color = color
        self.scale = scale

        if background is None:
            background = "ffffff"

        self.background = color._parse_color(background)

    def _render(self, width, height):

        domain = get_real_domain(width, height, scale=self.scale)
        drawable = Drawable(domain, self.shape, self.color)

        dimensions = (height, width, len(self.background))
        image_data = np.full(dimensions, self.background, dtype=np.uint8)

        return render_drawable(drawable, image_data)



import pytest
from unittest import TestCase

from stylo.shape import Circle
from stylo.color import FillColor
from stylo.image import LayeredImage
from stylo.testing.image import BaseImageTest


@pytest.mark.image
class TestLayeredImage(TestCase, BaseImageTest):
    """Tests for the :code:`LayeredImage` object."""

    def setUp(self):
        circle, black = Circle(0, 0, 0.9), FillColor()

        image = LayeredImage()
        image.add_layer(circle, black)

        self.image = image



import re
import struct
from abc import ABC, abstractmethod


rgb8_pattern = re.compile(r"\A[a-fA-F0-9]{6}\Z")


class ColorSpace(ABC):
    @staticmethod
    @abstractmethod
    def parse(string):
        """Parse the color from its representation."""
        pass


class RGB8(ColorSpace):
    """Represents the 8-bit RGB colorspace."""

    @staticmethod
    def parse(string):

        if rgb8_pattern.match(string):
            return struct.unpack("BBB", bytes.fromhex(string))

        raise ValueError("String does not represent a valid color.")



"""This module contains functions that automate the process of writing base test classes
for :code:`Domain` objects.

"""
import pytest
import numpy as np
from hypothesis import given

from stylo.testing.strategies import dimension


DOMAIN_TEST_DOCSTRING = """
A base class to use when testing {0} implementations.

When writing your test case declare it as follows:

.. code-block:: python

   from unittest import TestCase
   from stylo.testing.domain import Base{0}Test

   class TestMyDomain(TestCase, Base{0}Test):
       ...


.. note::

   The order in which you include these classes is *very* important.

You also need to define a :code:`setUp()` method to set the :code:`domain` attribute to
be an instance of your class.

.. code-block:: python

   def setUp(self):
       self.domain = MyDomain()

If you follow these steps in addition to the test cases you write checking the
specifics of your class it will also be checked to see that it conforms to the behavior
that is expected from every {0} instance.
"""

PARAMETER_READ_ONLY_DOCSTRING = """Ensure that the :code:`{0}` property is read only."""


PARAMETER_DOCSTRING = """
Ensure that the :code:`{0}` property works as expected.

The :code:`{0}` property should:

- Return a function in width and height.
- When called it should return a numpy array.
- The array should have shape :code:`(height, width)`
"""


def parameter_read_only_check(name):
    """Given the name of a parameter to check, make sure that the corresponding
    attribute is read-only.

    :param name: The name of the parameter
    :type name: str
    """

    def check(self):
        with pytest.raises(AttributeError) as err:
            self.domain.__setattr__(name, 2)

        self.assertIn("can't set attribute", str(err.value))

    check.__doc__ = PARAMETER_READ_ONLY_DOCSTRING.format(name)
    return check


def parameter_check(name):
    """Given the name of a parameter to check, make sure that the corresponding
    attribute returns a function that works as expected.

    :param name: The name of the parameter
    :type name: str
    """
    context = given(width=dimension, height=dimension)

    def check(self, width, height):

        param = self.domain.__getattribute__(name)
        self.assertTrue(
            callable(param), "The {} property should return a function".format(name)
        )

        ps = param(width, height)
        self.assertTrue(
            isinstance(ps, (np.ndarray,)), "The function should return a numpy array"
        )

        self.assertEqual(
            (height, width),
            ps.shape,
            "The resulting array should have shape (height, width)",
        )

    check.__doc__ = PARAMETER_DOCSTRING.format(name)
    return context(check)


def define_domain_test(domain):
    """Given a domain definition, write the test class that will automatically verify
    that a given subclass follows the expected interface.
    """
    test_name = "Base" + domain.__name__ + "Test"

    attributes = {"__doc__": DOMAIN_TEST_DOCSTRING.format(domain.__name__)}

    for param in domain._parameters:
        read_only = "test_parameter_{}_read_only".format(param)
        p = "test_parameter_{}".format(param)

        attributes[read_only] = parameter_read_only_check(param)
        attributes[p] = parameter_check(param)

    return type(test_name, (), attributes)



from .transform import RealDomainTransform, define_transform  # noqa: F401

from .rotation import rotate  # noqa: F401
from .shear import horizontal_shear, vertical_shear  # noqa: F401
from .translation import translate  # noqa: F401
